import base64
import binascii
import collections
import json
import multiprocessing
import time
import traceback
import uuid

import flask

# import smqtk.algorithms
from smqtk.algorithms import (
    get_classifier_impls,
    get_descriptor_generator_impls,
    get_nn_index_impls,
    get_relevancy_index_impls,
    SupervisedClassifier,
)
from smqtk.iqr import (
    iqr_controller,
    iqr_session,
)
from smqtk.representation import (
    ClassificationElementFactory,
    DescriptorElementFactory,
    get_descriptor_index_impls,
)
from smqtk.representation.data_element.memory_element import DataMemoryElement
from smqtk.utils import (
    merge_dict,
    plugin,
)
from smqtk.web import SmqtkWebApp


def new_uuid():
    return str(uuid.uuid1(clock_seq=int(time.time() * 1000000)))\
        .replace('-', '')


def make_response_json(message, **params):
    r = {
        "message": message,
        "time": {
            "unix": time.time(),
            "utc": time.asctime(time.gmtime()),
        }
    }
    merge_dict(r, params)
    return flask.jsonify(**r)


# Get expected JSON decode exception.
#
# Flask can use one of two potential JSON parsing libraries: simplejson or
# json.  simplejson has a specific exception for decoding errors while json
# just raises a ValueError.
#
# noinspection PyProtectedMember
if hasattr(flask.json._json, 'JSONDecodeError'):
    # noinspection PyProtectedMember
    JSON_DECODE_EXCEPTION = getattr(flask.json._json, 'JSONDecodeError')
else:
    JSON_DECODE_EXCEPTION = ValueError


def parse_hashable_json_list(json_str):
    """
    Parse and check input string, looking for a JSON list of hashable values.

    :param json_str: String to parse and check.
    :type json_str: str

    :raises ValueError: Expected value check failed.

    :return: List of hashable-type values.
    :rtype: list[collections.Hashable]

    """
    try:
        v_list = flask.json.loads(json_str)
    except JSON_DECODE_EXCEPTION as ex:
        raise ValueError("JSON parsing error: %s" % str(ex))
    if not isinstance(v_list, list):
        raise ValueError("JSON provided is not a list.")
    # Should not be an empty list.
    elif not v_list:
        raise ValueError("JSON list is empty.")
    # Contents of list should be numeric or string values.
    elif not all(isinstance(el, collections.Hashable)
                 for el in v_list):
        raise ValueError("Not all JSON list parts were hashable values.")
    return v_list


class IqrService (SmqtkWebApp):
    """
    Configuration Notes
    -------------------
    ``descriptor_index`` will currently be configured twice: once for the
    global index and once for the nearest neighbors index. These will probably
    be the set to the same index. In more detail, the global descriptor index
    is used when the "refine" endpoint is given descriptor UUIDs
    """

    @classmethod
    def is_usable(cls):
        return True

    @classmethod
    def get_default_config(cls):
        c = super(IqrService, cls).get_default_config()

        c_rel_index = plugin.make_config(
            get_relevancy_index_impls()
        )
        merge_dict(c_rel_index, iqr_session.DFLT_REL_INDEX_CONFIG)

        merge_dict(c, {
            "iqr_service": {

                "session_control": {
                    "positive_seed_neighbors": 500,
                    "session_expiration": {
                        "enabled": False,
                        "check_interval_seconds": 30,
                        "session_timeout": 3600,
                    }
                },

                "plugin_notes": {
                    "relevancy_index_config":
                        "The relevancy index config provided should not have "
                        "persistent storage configured as it will be used in "
                        "such a way that instances are created, built and "
                        "destroyed often.",
                    "descriptor_factory":
                        "What descriptor element factory to use when asked to "
                        "compute a descriptor on data.",
                    "descriptor_generator":
                        "Descriptor generation algorithm to use when "
                        "requested to describe data.",
                    "descriptor_index":
                        "This is the index from which given positive and "
                        "negative example descriptors are retrieved from. "
                        "Not used for nearest neighbor querying. "
                        "This index must contain all descriptors that could "
                        "possibly be used as positive/negative examples and "
                        "updated accordingly.",
                    "neighbor_index":
                        "This is the neighbor index to pull initial near-"
                        "positive descriptors from.",
                    "classifier_config":
                        "The configuration to use for training and using "
                        "classifiers for the /classifier endpoint. "
                        "When configuring a classifier for use, don't fill "
                        "out model persistence values as many classifiers "
                        "may be created and thrown away during this service's "
                        "operation.",
                    "classification_factory":
                        "Selection of the backend in which classifications "
                        "are stored. The in-memory version is recommended "
                        "because normal caching mechanisms will not account "
                        "for the variety of classifiers that can potentially "
                        "be created via this utility.",
                },

                "plugins": {
                    "relevancy_index_config": c_rel_index,
                    "descriptor_factory":
                        DescriptorElementFactory.get_default_config(),
                    "descriptor_generator": plugin.make_config(
                        get_descriptor_generator_impls()
                    ),
                    "descriptor_index": plugin.make_config(
                        get_descriptor_index_impls()
                    ),
                    "neighbor_index":
                        plugin.make_config(get_nn_index_impls()),
                    "classifier_config":
                        plugin.make_config(get_classifier_impls()),
                    "classification_factory":
                        ClassificationElementFactory.get_default_config(),
                },

            }
        })
        return c

    def __init__(self, json_config):
        super(IqrService, self).__init__(json_config)
        sc_config = json_config['iqr_service']['session_control']

        # Initialize from config
        self.positive_seed_neighbors = sc_config['positive_seed_neighbors']
        self.classifier_config = \
            json_config['iqr_service']['plugins']['classifier_config']
        self.classification_factory = \
            ClassificationElementFactory.from_config(
                json_config['iqr_service']['plugins']['classification_factory']
            )

        self.descriptor_factory = DescriptorElementFactory.from_config(
            json_config['iqr_service']['plugins']['descriptor_factory']
        )

        #: :type: smqtk.algorithms.DescriptorGenerator
        self.descriptor_generator = plugin.from_plugin_config(
            json_config['iqr_service']['plugins']['descriptor_generator'],
            get_descriptor_generator_impls(),
        )

        #: :type: smqtk.representation.DescriptorIndex
        self.descriptor_index = plugin.from_plugin_config(
            json_config['iqr_service']['plugins']['descriptor_index'],
            get_descriptor_index_impls(),
        )

        #: :type: smqtk.algorithms.NearestNeighborsIndex
        self.neighbor_index = plugin.from_plugin_config(
            json_config['iqr_service']['plugins']['neighbor_index'],
            get_nn_index_impls(),
        )
        self.neighbor_index_lock = multiprocessing.RLock()

        self.rel_index_config = \
            json_config['iqr_service']['plugins']['relevancy_index_config']

        # Record of trained classifiers for a session. Session classifier
        # modifications locked under the parent session's global lock.
        #: :type: dict[collections.Hashable, SupervisedClassifier | None]
        self.session_classifiers = {}
        # Control for knowing when a new classifier should be trained for a
        # session (True == train new classifier). Modification for specific
        # sessions under parent session's lock.
        #: :type: dict[collections.Hashable, bool]
        self.session_classifier_dirty = {}

        def session_expire_callback(session):
            """
            :type session: smqtk.iqr.IqrSession
            """
            with session:
                self._log.debug("Removing session %s classifier", session.uuid)
                del self.session_classifiers[session.uuid]
                del self.session_classifier_dirty[session.uuid]

        self.controller = iqr_controller.IqrController(
            sc_config['session_expiration']['enabled'],
            sc_config['session_expiration']['check_interval_seconds'],
            session_expire_callback
        )
        self.session_timeout = \
            sc_config['session_expiration']['session_timeout']

        self.add_routes()

    def add_routes(self):
        """
        Setup Flask URL rules.
        """
        self.add_url_rule('/is_ready',
                          view_func=self.is_ready,
                          methods=['GET'])
        self.add_url_rule('/add_descriptor_from_data',
                          view_func=self.add_descriptor_from_data,
                          methods=['POST'])
        # TODO: Potentially other add_descriptor_from_* variants that expect
        #       other forms of input besides base64, like arbitrary URIs (to
        #       use from_uri factory function).
        self.add_url_rule('/nn_index',
                          view_func=self.get_nn_index_status,
                          methods=['GET'])
        self.add_url_rule('/nn_index',
                          view_func=self.update_nn_index,
                          methods=['POST'])
        self.add_url_rule('/nn_index',
                          view_func=self.remove_from_nn_index,
                          methods=['DELETE'])
        self.add_url_rule('/data_nearest_neighbors',
                          view_func=self.data_nearest_neighbors,
                          methods=['POST'])
        self.add_url_rule('/uid_nearest_neighbors',
                          view_func=self.uid_nearest_neighbors,
                          methods=['POST'])
        self.add_url_rule('/session_ids',
                          view_func=self.get_sessions_ids,
                          methods=['GET'])
        self.add_url_rule('/session',
                          view_func=self.get_session_info,
                          methods=['GET'])
        self.add_url_rule('/session',
                          view_func=self.init_session,
                          methods=['POST'])
        self.add_url_rule('/session',
                          view_func=self.reset_session,
                          methods=['PUT'])
        self.add_url_rule('/session',
                          view_func=self.clean_session,
                          methods=['DELETE'])
        self.add_url_rule('/add_external_pos',
                          view_func=self.add_external_positive,
                          methods=['POST'])
        self.add_url_rule('/add_external_neg',
                          view_func=self.add_external_negative,
                          methods=['POST'])
        self.add_url_rule('/adjudicate',
                          view_func=self.get_adjudication,
                          methods=['GET'])
        self.add_url_rule('/adjudicate',
                          view_func=self.adjudicate,
                          methods=['POST'])
        self.add_url_rule('/initialize',
                          view_func=self.initialize,
                          methods=['POST'])
        self.add_url_rule('/refine',
                          view_func=self.refine,
                          methods=['POST'])
        self.add_url_rule('/num_results',
                          view_func=self.num_results,
                          methods=['GET'])
        self.add_url_rule('/get_results',
                          view_func=self.get_results,
                          methods=['GET'])
        self.add_url_rule('/classify',
                          view_func=self.classify,
                          methods=['GET'])
        self.add_url_rule('/state',
                          view_func=self.get_iqr_state,
                          methods=['GET'])
        self.add_url_rule('/state',
                          view_func=self.set_iqr_state,
                          methods=['PUT'])

    def describe_base64_data(self, b64, content_type):
        """
        Compute and return the descriptor element for the given base64 data.

        The given data bytes are not retained.

        :param b64: Base64 data string.
        :type b64: str

        :param content_type: Data content type.
        :type content_type: str

        :raises TypeError: Failed to parse base64 data.

        :return: Computed descriptor element.
        :rtype: smqtk.representation.DescriptorElement
        """
        de = DataMemoryElement.from_base64(b64, content_type)
        return self.descriptor_generator.compute_descriptor(
            de, self.descriptor_factory
        )

    # GET /is_ready
    # noinspection PyMethodMayBeStatic
    def is_ready(self):
        """
        Simple function that returns True, indicating that the server is
        active.
        """
        return make_response_json("Yes, I'm alive."), 200

    # POST /add_descriptor_from_data
    def add_descriptor_from_data(self):
        """
        Add the description of the given base64 data with content type to the
        descriptor set.

        Accept base64 data (with content type), describe it via the configured
        descriptor generator and add the resulting descriptor element to the
        configured descriptor index.

        Form Arguments:
            data_b64
                Base64-encoded input binary data to describe via
                DescriptorGenerator.  This must be of a content type accepted
                by the configured DescriptorGenerator.
            content_type
                Input data content mimetype string.

        JSON return object:
            uid
                UID of the descriptor element generated from input data
                description.  This should be equivalent to the SHA1 checksum of
                the input data.
            size
                New size (integer) of the descriptor set that has been updated
                (NOT the same as the nearest-neighbor index).

        """
        data_b64 = flask.request.form.get('data_b64', None)
        content_type = flask.request.form.get('content_type', None)
        if not data_b64:
            return make_response_json("No or empty base64 data provided."), 400
        if not content_type:
            return make_response_json("No data mimetype provided."), 400

        try:
            descriptor = self.describe_base64_data(data_b64, content_type)
        except (TypeError, binascii.Error) as e:
            if str(e) == "Incorrect padding":
                return make_response_json("Failed to parse base64 data."), 400
            # In case some other exception is raised, actually a server error.
            raise
        # Concurrent updating of descriptor set should be handled by underlying
        # implementation.
        self.descriptor_index.add_descriptor(descriptor)
        return make_response_json("Success",
                                  uid=descriptor.uuid(),
                                  size=self.descriptor_index.count()), 201

    # GET /nn_index
    def get_nn_index_status(self):
        """
        Get status/state information about the nearest-neighbor index.

        Status code 200 on success, JSON return object: {
            ...,
            // Size of the nearest-neighbor index.
            index_size=<int>
        }
        """
        with self.neighbor_index_lock:
            return (
                make_response_json("Success",
                                   index_size=self.neighbor_index.count()),
                200
            )

    # POST /nn_index
    def update_nn_index(self):
        """
        Tell the configured nearest-neighbor-index instance to update with the
        descriptors associated with the provided list of UIDs.

        This is a critical operation on the index so this method can only be
        invoked once at a time (other concurrent will block until previous
        calls have finished).

        Form Arguments:
            descriptor_uids
                JSON list of UID strings.  If one or more UIDs do not match
                descriptors in our current descriptor-set we return an error
                message.

        JSON return object:
            message
                Success string
            descriptor_uids
                List of UIDs the neighbor index was updated with.  This should
                be congruent with the list provided.
            index_size
                New size of the nearest-neighbors index.

        """
        descr_uid_str = flask.request.form.get('descriptor_uids', None)
        if not descr_uid_str:  # empty string or None
            return make_response_json("No descriptor UID JSON provided."), 400

        # Load and check JSON input.
        try:
            descr_uid_list = parse_hashable_json_list(descr_uid_str)
        except ValueError as ex:
            return make_response_json("%s" % str(ex)), 400

        with self.neighbor_index_lock:
            try:
                # KeyError may not occur until returned iterator is iterated.
                descr_elems = \
                    self.descriptor_index.get_many_descriptors(descr_uid_list)
                self.neighbor_index.update_index(descr_elems)
            except KeyError:
                # Some UIDs are not present in the current index.  Isolate
                # which UIDs are not contained.
                uids_not_ingested = []
                for uid in descr_uid_list:
                    if not self.descriptor_index.has_descriptor(uid):
                        uids_not_ingested.append(uid)
                return make_response_json("Some provided UIDs do not exist in "
                                          "the current index.",
                                          bad_uids=uids_not_ingested), 400

            return (
                make_response_json("Success",
                                   descriptor_uids=descr_uid_list,
                                   index_size=self.neighbor_index.count()),
                200
            )

    # DELETE /nn_index
    def remove_from_nn_index(self):
        """
        Remove descriptors from the nearest-neighbors index given their UIDs.

        Receive one or more descriptor UIDs, that exist in the NN-index, that
        are to be removed from the NN-index.  This DOES NOT remove elements
        from the global descriptor set.

        This is a critical operation on the index so this method can only be
        invoked once at a time (other concurrent will block until previous
        calls have finished).

        Form Arguments:
            descriptor_uids
                JSON list of descriptor UIDs to remove from the nearest-
                neighbor index.  These UIDs must be present in the index,
                otherwise an 404 error is returned.

        Status code 200 on success, JSON return object: {
            ...,

            // List of UID values removed from the index.
            descriptor_uids=<list[str]>,

            // New size of the nearest-neighbors index.
            index_size=<int>
        }

        """
        descr_uid_str = flask.request.values.get('descriptor_uids', None)
        if not descr_uid_str:  # empty string or None
            return make_response_json("No descriptor UID JSON provided."), 400

        # Load and check JSON input
        try:
            descr_uid_list = parse_hashable_json_list(descr_uid_str)
        except ValueError as ex:
            return make_response_json("%s" % str(ex)), 400

        with self.neighbor_index_lock:
            try:
                # empty list already checked for in above try-catch, so we
                # should never see a ValueError here.  KeyError still possible.
                self.neighbor_index.remove_from_index(descr_uid_list)
            except KeyError as ex:
                return make_response_json("Some provided UIDs do not exist in "
                                          "the current index.",
                                          bad_uids=ex.args), 400

            return (
                make_response_json("Success",
                                   descriptor_uids=descr_uid_list,
                                   index_size=self.neighbor_index.count()),
                200
            )

    # POST /data_nearest_neighbors
    def data_nearest_neighbors(self):
        """
        Take in data in base64 encoding with a mimetype and find its 'k'
        nearest neighbors according to the current index, including their
        distance values (metric determined by nearest-neighbors-index algorithm
        configuration).

        This endpoint does not need a session ID due to the
        nearest-neighbor-index being a shared resource across IQR sessions.

        Form Arguments:
            data_b64
                Base64-encoded input binary data to describe via
                DescriptorGenerator.  This must be of a content type accepted
                by the configured DescriptorGenerator.
            content_type
                Input data content mimetype string.
            k
                Integer number of nearest neighbor descriptor UIDs to return
                along with their distances.

        JSON return object:
            neighbor_uids
                Ordered list of neighbor UID values. Index 0 represents the
                closest neighbor while the last index represents the farthest
                neighbor.  Parallel in relationship to `neighbor_dists`.
            neighbor_dists
                Ordered list of neighbor distance values. Index 0 represents
                the closest neighbor while the last index represents the
                farthest neighbor.  Parallel in relationship to
                'neighbor_uids`.

        """
        data_b64 = flask.request.form.get('data_b64', None)
        content_type = flask.request.form.get('content_type', None)
        k_str = flask.request.form.get('k', None)
        if not data_b64:
            return make_response_json("No or empty base64 data provided."), 400
        if not content_type:
            return make_response_json("No data mimetype provided."), 400
        if not k_str:
            return make_response_json("No 'k' value provided."), 400

        try:
            k = int(k_str)
        except ValueError as ex:
            return make_response_json("Failed to convert 'k' argument to an "
                                      "integer: %s" % str(ex)), 400

        try:
            descriptor = self.describe_base64_data(data_b64, content_type)
        except (TypeError, binascii.Error) as e:
            return make_response_json("Failed to parse base64 data: {}"
                                      .format(str(e))), 400

        n_elems, n_dists = self.neighbor_index.nn(descriptor, k)
        return make_response_json("Success",
                                  neighbor_uids=[e.uuid() for e in n_elems],
                                  neighbor_dists=[d for d in n_dists]), 200

    # POST /uid_nearest_neighbors
    def uid_nearest_neighbors(self):
        """
        Take in the UID that matches an ingested descriptor and find that
        descriptor's 'k' nearest neighbors according to the current index,
        including their distance values (metric determined by
        nearest-neighbors-index algorithm configuration).

        This endpoint does not need a session ID due to the
        nearest-neighbor-index being a shared resource across IQR sessions.

        This endpoint can be more advantageous compared the
        `data_nearest_neighbors` endpoint if you know a descriptor has already
        been ingested (via `add_descriptor_from_data` or otherwise) as a
        potentially new descriptor does not have to be computed.

        Form Arguments:
            uid
                UID of the descriptor to get the nearest neighbors for.  This
                should also match the SHA1 checksum of the data being
                described.
            k
                Integer number of nearest neighbor descriptor UIDs to return
                along with their distances.

        JSON return object:
            neighbor_uids
                Ordered list of neighbor UID values. Index 0 represents the
                closest neighbor while the last index represents the farthest
                neighbor.  Parallel in relationship to `neighbor_dists`.
            neighbor_dists
                Ordered list of neighbor distance values. Index 0 represents
                the closest neighbor while the last index represents the
                farthest neighbor.  Parallel in relationship to
                'neighbor_uids`.

        """
        uid = flask.request.form.get('uid', None)
        k_str = flask.request.form.get('k', None)
        if not uid:
            return make_response_json("No UID provided."), 400
        if not k_str:
            return make_response_json("No 'k' value provided."), 400

        try:
            k = int(k_str)
        except ValueError as ex:
            return make_response_json("Failed to convert 'k' argument to an "
                                      "integer: %s" % str(ex)), 400

        try:
            descriptor = self.descriptor_index.get_descriptor(uid)
        except KeyError:
            return make_response_json("Failed to get descriptor for UID %s"
                                      % uid), 400

        n_elems, n_dists = self.neighbor_index.nn(descriptor, k)
        return make_response_json("Success",
                                  neighbor_uids=[e.uuid() for e in n_elems],
                                  neighbor_dists=[d for d in n_dists]), 200

    # GET /session_ids
    def get_sessions_ids(self):
        """
        Get the list of current, active session IDs.
        """
        session_uuids = self.controller.session_uuids()
        return make_response_json("Current session UUID values",
                                  session_uuids=session_uuids), 200

    # GET /session
    def get_session_info(self):
        """
        Get a JSON return with session state information.

        URL Arguments:
            sid
                ID of the session.
        """
        sid = flask.request.args.get('sid', None)
        if sid is None:
            return make_response_json("No session id (sid) provided"), 400

        with self.controller:
            if not self.controller.has_session_uuid(sid):
                return make_response_json("session id '%s' not found" % sid,
                                          sid=sid), 404
            iqrs = self.controller.get_session(sid)
            iqrs.lock.acquire()  # lock BEFORE releasing controller

        try:
            uuids_pos = [d.uuid() for d in iqrs.positive_descriptors]
            uuids_pos_external = [d.uuid() for d in
                                  iqrs.external_positive_descriptors]
            uuids_neg = [d.uuid() for d in iqrs.negative_descriptors]
            uuids_neg_external = [d.uuid() for d in
                                  iqrs.external_negative_descriptors]
            wi_count = iqrs.working_index.count()

        finally:
            iqrs.lock.release()

        return make_response_json("Session '%s' info" % sid,
                                  sid=sid,
                                  uuids_pos=uuids_pos,
                                  uuids_neg=uuids_neg,
                                  uuids_pos_ext=uuids_pos_external,
                                  uuids_neg_ext=uuids_neg_external,
                                  wi_count=wi_count), 200

    # POST /session
    def init_session(self):
        """
        Initialize a new session in the controller.

        Form args:
            sid
                Optional UUID string specification of session to initialize. If
                one is not provided, we create one here.

        """
        sid = flask.request.form.get('sid', None)
        if sid is None:
            sid = new_uuid()

        if self.controller.has_session_uuid(sid):
            return make_response_json(
                "Session with id '%s' already exists" % sid,
                sid=sid,
            ), 409  # CONFLICT

        iqrs = iqr_session.IqrSession(self.positive_seed_neighbors,
                                      self.rel_index_config,
                                      sid)
        with self.controller:
            with iqrs:  # because classifier maps locked by session
                self.controller.add_session(iqrs, self.session_timeout)
                self.session_classifiers[sid] = None
                self.session_classifier_dirty[sid] = True

        return make_response_json("Created new session with ID '%s'" % sid,
                                  sid=sid), 201  # CREATED

    # PUT /session
    def reset_session(self):
        """
        Reset an existing session. This does not remove the session, so actions
        may

        Form args:
            sid
                UUID (string) of the session to reset.

        """
        sid = flask.request.form.get('sid', None)

        if sid is None:
            return make_response_json("No session id (sid) provided"), 400

        with self.controller:
            if not self.controller.has_session_uuid(sid):
                return make_response_json("session id '%s' not found" % sid,
                                          sid=sid), 404
            iqrs = self.controller.get_session(sid)
            iqrs.lock.acquire()  # lock BEFORE releasing controller

        try:
            iqrs.reset()
            self.session_classifiers[sid] = None
            self.session_classifier_dirty[sid] = True

        finally:
            iqrs.lock.release()

        return make_response_json("Reset IQR session '%s'" % sid,
                                  sid=sid), 200

    # DELETE /session
    def clean_session(self):
        """
        Clean resources associated with the session of the given UUID.

        Similar to session resetting, but this also removed the session
        resource causing future references without another initialize to return
        errors.

        Form args:
            sid
                UUID of the session to clean.

        """
        sid = flask.request.form.get('sid', None)

        if sid is None:
            return make_response_json("No session id (sid) provided"), 400

        with self.controller:
            if not self.controller.has_session_uuid(sid):
                return make_response_json("session id '%s' not found" % sid,
                                          sid=sid), 404
            with self.controller.get_session(sid) as iqrs:
                iqrs.reset()
                del self.session_classifiers[sid]
                del self.session_classifier_dirty[sid]
            self.controller.remove_session(sid)
        return make_response_json("Cleaned session resources for '%s'" % sid,
                                  sid=sid), 200

    # POST /add_external_pos
    def add_external_positive(self):
        """
        Describe the given data and store as a positive example from external
        data.

        Form args:
            sid
                The id of the session to add the generated descriptor to.
            base64
                The base64 byes of the data. This should use the standard or
                URL-safe alphabet as the python ``base64.urlsafe_b64decode``
                module function would expect (handles either alphabet).
            content_type
                The mimetype of the bytes given.

        A return JSON along with a code 201 means a descriptor was successfully
        computed and added to the session external positives. The returned JSON
        includes a reference to the UUID of the descriptor computed under the
        ``descr_uuid`` key.

        """
        sid = flask.request.form.get('sid', None)
        data_base64 = flask.request.form.get('base64', None)
        data_content_type = flask.request.form.get('content_type', None)

        if sid is None:
            return make_response_json("No session id (sid) provided"), 400
        if not data_base64:
            return make_response_json("No or empty base64 data provided."), 400
        if not data_content_type:
            return make_response_json("No data mimetype provided."), 400

        descriptor = self.describe_base64_data(data_base64, data_content_type)

        with self.controller:
            if not self.controller.has_session_uuid(sid):
                return make_response_json("session id '%s' not found" % sid,
                                          sid=sid), 404
            iqrs = self.controller.get_session(sid)
            iqrs.lock.acquire()  # lock BEFORE releasing controller

        try:
            iqrs.external_descriptors(positive=[descriptor])

        finally:
            iqrs.lock.release()

        return make_response_json("Success", descr_uuid=descriptor.uuid()), 201

    # POST /add_external_neg
    def add_external_negative(self):
        """
        Describe the given data and store as a negative example from external
        data.

        Form args:
            sid
                The id of the session to add the generated descriptor to.
            base64
                The base64 byes of the data. This should use the standard or
                URL-safe alphabet as the python ``base64.urlsafe_b64decode``
                module function would expect (handles either alphabet).
            content_type
                The mimetype of the bytes given.

        A return JSON along with a code 201 means a descriptor was successfully
        computed and added to the session external negatives. The returned JSON
        includes a reference to the UUID of the descriptor computed under the
        ``descr_uuid`` key.

        """
        sid = flask.request.form.get('sid', None)
        data_base64 = flask.request.form.get('base64', None)
        data_content_type = flask.request.form.get('content_type', None)

        if sid is None:
            return make_response_json("No session id (sid) provided"), 400
        if not data_base64:
            return make_response_json("No or empty base64 data provided."), 400
        if not data_content_type:
            return make_response_json("No data mimetype provided."), 400

        descriptor = self.describe_base64_data(data_base64,
                                               data_content_type)

        with self.controller:
            if not self.controller.has_session_uuid(sid):
                return make_response_json("session id '%s' not found" % sid,
                                          sid=sid), 404
            iqrs = self.controller.get_session(sid)
            iqrs.lock.acquire()  # lock BEFORE releasing controller

        try:
            iqrs.external_descriptors(negative=[descriptor])

        finally:
            iqrs.lock.release()

        return make_response_json("Success",
                                  descr_uuid=descriptor.uuid()), 201

    # GET /adjudicate
    def get_adjudication(self):
        """
        Get the adjudication state of a descriptor given its UID.

        Form args:
            sid
                Session Id.
            uid
                Query descriptor UID.
        """
        sid = flask.request.args.get('sid', None)
        uid = flask.request.args.get('uid', None)

        if sid is None:
            return make_response_json("No session id (sid) provided"), 400
        elif uid is None:
            return make_response_json("No descriptor uid provided"), 400

        with self.controller:
            if not self.controller.has_session_uuid(sid):
                return make_response_json("session id '%s' not found" % sid,
                                          sid=sid), 404
            iqrs = self.controller.get_session(sid)
            iqrs.lock.acquire()  # lock BEFORE releasing controller

        try:
            all_pos = (iqrs.external_positive_descriptors |
                       iqrs.positive_descriptors)
            all_neg = (iqrs.external_negative_descriptors |
                       iqrs.negative_descriptors)

        finally:
            iqrs.lock.release()

        is_pos = uid in {d.uuid() for d in all_pos}
        is_neg = uid in {d.uuid() for d in all_neg}

        if is_pos and is_neg:
            return make_response_json("UID slotted as both positive and "
                                      "negative?"), 500

        return make_response_json("%s descriptor adjudication" % uid,
                                  is_pos=is_pos, is_neg=is_neg), 200

    # POST /adjudicate
    def adjudicate(self):
        """
        Incrementally update internal adjudication state given new positives
        and negatives, and optionally IDs for descriptors now marked neutral.

        If the same UUID is present in both positive and negative sets, they
        cancel each other out (remains neutral).

        Descriptor uuids that may be provided must be available in the
        configured descriptor index.

        Form Args:
            sid
                UUID of session to interact with.
            pos
                List of descriptor UUIDs that should be marked positive.
            neg
                List of descriptor UUIDs that should be marked negative.
            neutral
                List of descriptor UUIDs that should not be marked positive or
                negative. If a UUID present in this list and in either pos or
                neg will be considered neutral.

        """
        sid = flask.request.form.get('sid', None)
        pos_uuids = flask.request.form.get('pos', '[]')
        neg_uuids = flask.request.form.get('neg', '[]')
        neu_uuids = flask.request.form.get('neutral', '[]')

        if sid is None:
            return make_response_json("No session id (sid) provided"), 400

        pos_uuids = set(json.loads(pos_uuids))
        neg_uuids = set(json.loads(neg_uuids))
        neu_uuids = set(json.loads(neu_uuids))

        with self.controller:
            if not self.controller.has_session_uuid(sid):
                return make_response_json("session id '%s' not found" % sid,
                                          sid=sid), 404
            iqrs = self.controller.get_session(sid)
            iqrs.lock.acquire()  # lock BEFORE releasing controller

        try:
            self._log.debug("Getting the descriptors for UUIDs")
            # get_many_descriptors can raise KeyError
            pos_d = set(
                self.descriptor_index.get_many_descriptors(pos_uuids)
            )
            neg_d = set(
                self.descriptor_index.get_many_descriptors(neg_uuids)
            )
            neu_d = set(
                self.descriptor_index.get_many_descriptors(neu_uuids)
            )

            # Record previous pos/neg descriptors for determining if an
            # existing classifier is dirty after this adjudication.
            orig_pos = set(iqrs.positive_descriptors)
            orig_neg = set(iqrs.negative_descriptors)

            self._log.debug("[%s] Adjudicating", sid)
            iqrs.adjudicate(pos_d, neg_d, neu_d, neu_d)

            # Flag classifier as dirty if change in pos/neg sets
            diff_pos = \
                iqrs.positive_descriptors.symmetric_difference(orig_pos)
            diff_neg = \
                iqrs.negative_descriptors.symmetric_difference(orig_neg)
            if diff_pos or diff_neg:
                self._log.debug("[%s] session Classifier dirty", sid)
                self.session_classifier_dirty[sid] = True

        except KeyError as ex:
            err_uuid = str(ex)
            self._log.warn(traceback.format_exc())
            return make_response_json(
                "Descriptor UUID '%s' cannot be found in the "
                "configured descriptor index."
                % err_uuid,
                sid=sid,
                uuid=err_uuid,
            ), 404

        finally:
            iqrs.lock.release()

        return make_response_json(
            "Finished adjudication",
            sid=sid,
        ), 200

    # POST /initialize
    def initialize(self):
        """
        Update the working index based on the currently positive examples and
        adjudications.

        Form Arguments:
            sid
                Id of the session to update.
        """
        sid = flask.request.form.get('sid', None)
        if sid is None:
            return make_response_json("No session id (sid) provided"), 400

        with self.controller:
            if not self.controller.has_session_uuid(sid):
                return make_response_json("session id '%s' not found" % sid,
                                          sid=sid,
                                          success=False), 404
            iqrs = self.controller.get_session(sid)
            iqrs.lock.acquire()  # lock BEFORE releasing controller

        try:
            iqrs.update_working_index(self.neighbor_index)
        except RuntimeError as ex:
            if "No positive descriptors to query" in str(ex):
                return make_response_json("Failed to initialize, no positive "
                                          "descriptors to query",
                                          sid=sid, success=False), 200
            else:
                raise
        finally:
            iqrs.lock.release()

        return make_response_json("Success", sid=sid, success=True), 200

    # POST /refine
    def refine(self):
        """
        (Re)Create ranking of working index content by order of relevance to
        examples and adjudications.

        Form args:
            sid
                Id of the session to use.

        """
        sid = flask.request.form.get('sid', None)

        if sid is None:
            return make_response_json("No session id (sid) provided"), 400

        with self.controller:
            if not self.controller.has_session_uuid(sid):
                return make_response_json("session id %s not found" % sid,
                                          sid=sid), 404
            iqrs = self.controller.get_session(sid)
            iqrs.lock.acquire()  # lock BEFORE releasing controller

        try:
            self._log.info("[%s] Refining", sid)
            iqrs.refine()

        finally:
            iqrs.lock.release()

        return make_response_json("Refine complete", sid=sid), 201

    # GET /num_results
    def num_results(self):
        """
        Get the total number of results in the ranking.

        URI Args:
            sid
                UUID of the session to use

        """
        sid = flask.request.args.get('sid', None)

        if sid is None:
            return make_response_json("No session id (sid) provided"), 400

        with self.controller:
            if not self.controller.has_session_uuid(sid):
                return make_response_json("session id '%s' not found" % sid,
                                          sid=sid), 404
            iqrs = self.controller.get_session(sid)
            iqrs.lock.acquire()  # lock BEFORE releasing controller

        try:
            if iqrs.results:
                size = len(iqrs.results)
            else:
                size = 0

        finally:
            iqrs.lock.release()

        return make_response_json("Currently %d results for session %s"
                                  % (size, sid),
                                  num_results=size,
                                  sid=sid), 200

    # GET /get_results
    def get_results(self):
        """
        Get the ordered ranking results between two index positions (inclusive,
        exclusive).

        This returns a results slice in the same way that python would handle a
        list slice.

        If the requested session has not been refined yet (no ranking), all
        result requests will be empty.

        URI Args:
            sid
                UUID of the session to use
            i
                Starting index (inclusive)
            j
                Ending index (exclusive)

        """
        sid = flask.request.args.get('sid', None)
        i = flask.request.args.get('i', None)
        j = flask.request.args.get('j', None)

        if sid is None:
            return make_response_json("No session id (sid) provided"), 400

        with self.controller:
            if not self.controller.has_session_uuid(sid):
                return make_response_json("session id '%s' not found" % sid,
                                          sid=sid), 404
            iqrs = self.controller.get_session(sid)
            iqrs.lock.acquire()  # lock BEFORE releasing controller

        try:
            if not iqrs.results:
                return make_response_json("Returning result pairs",
                                          i=0, j=0, total_results=0,
                                          results=[], sid=sid), 200

            num_results = len(iqrs.results)

            # int() can raise ValueError, catch
            i = 0 if i is None else int(i)
            j = num_results if j is None else int(j)

            # We ensured i, j are valid by this point
            r = [[d.uuid(), prob] for d, prob in iqrs.ordered_results()[i:j]]

        except ValueError:
            return make_response_json("Invalid bounds index value(s)"), 400

        finally:
            iqrs.lock.release()

        return make_response_json("Returning result pairs",
                                  i=i, j=j, total_results=num_results,
                                  results=r,
                                  sid=sid), 200

    # GET /classify
    def classify(self):
        """
        Given a refined session ID and some number of descriptor UUIDs, create
        a classifier according to the current state and classify the given
        descriptors adjudicated.

        This will fail if the session has not been given adjudications
        (refined) yet.

        URI Args:
            sid
                UUID of the session to utilize
            uuids
                List of descriptor UUIDs to classify. These UUIDs must
                associate to descriptors in the configured descriptor index.

        TODO: Optionally take in a list of JSON objects encoding base64 bytes
              and content type of raw data to describe and then classify, thus
              extending classification ability to arbitrary new data.

        """
        # Record clean/dirty status after making classifier/refining so we
        # don't train a new classifier when we don't have to.
        sid = flask.request.args.get('sid', None)
        uuids = flask.request.args.get('uuids', None)

        if sid is None:
            return make_response_json("No session id (sid) provided"), 400

        try:
            uuids = json.loads(uuids)
        except ValueError:
            return make_response_json(
                "Failed to decode uuids as json. Given '%s'"
                % uuids
            ), 400

        with self.controller:
            if not self.controller.has_session_uuid(sid):
                return make_response_json("session id '%s' not found" % sid,
                                          sid=sid), 404
            iqrs = self.controller.get_session(sid)
            iqrs.lock.acquire()  # lock BEFORE releasing controller

        try:
            all_pos = (iqrs.external_positive_descriptors |
                       iqrs.positive_descriptors)
            if not all_pos:
                return make_response_json(
                    "No positive labels in current session. Required for a "
                    "supervised classifier.",
                    sid=sid
                ), 400
            all_neg = (iqrs.external_negative_descriptors |
                       iqrs.negative_descriptors)
            if not all_neg:
                return make_response_json(
                    "No negative labels in current session. Required for a "
                    "supervised classifier.",
                    sid=sid
                ), 400

            # Get descriptor elements for classification
            # get_many_descriptors can raise KeyError
            descriptors = list(self.descriptor_index
                               .get_many_descriptors(uuids))

            classifier = self.session_classifiers.get(sid, None)

            pos_label = "positive"
            neg_label = "negative"
            if self.session_classifier_dirty[sid] or classifier is None:
                self._log.debug("Training new classifier for current "
                                "adjudication state...")

                #: :type: SupervisedClassifier
                classifier = plugin.from_plugin_config(
                    self.classifier_config,
                    get_classifier_impls(sub_interface=SupervisedClassifier)
                )
                classifier.train(
                    {pos_label: all_pos,
                     neg_label: all_neg}
                )

                self.session_classifiers[sid] = classifier
                self.session_classifier_dirty[sid] = False

            classifications = classifier.classify_async(
                descriptors, self.classification_factory,
                use_multiprocessing=True, ri=1.0
            )

            # Format output to be parallel lists of UUIDs input and
            # positive class classification scores.
            o_uuids = []
            o_proba = []
            for d in descriptors:
                o_uuids.append(d.uuid())
                o_proba.append(classifications[d][pos_label])

            assert uuids == o_uuids, \
                "Output UUID list is not congruent with INPUT list."

        except KeyError as ex:
            err_uuid = str(ex)
            self._log.warn(traceback.format_exc())
            return make_response_json(
                "Descriptor UUID '%s' cannot be found in the "
                "configured descriptor index."
                % err_uuid,
                sid=sid,
                uuid=err_uuid,
            ), 404

        finally:
            iqrs.lock.release()

        return make_response_json(
            "Finished classification",
            sid=sid,
            uuids=o_uuids,
            proba=o_proba,
        ), 200

    # TODO: Save/Export classifier model/state/configuration?

    # GET /state
    def get_iqr_state(self):
        """ [See api.rst]
        Create and return a binary package representing this IQR session's
        state.

        An IQR state is composed of the descriptor vectors, and their UUIDs,
        that were added from external sources, or were adjudicated, positive
        and negative.

        URL Arguments:
            sid
                Session ID to get the state of.

        Possible error code returns:
            400
                No session ID provided.
            404
                No session for the given ID.

        Success returns 200: {
            message = "Success"
            ...
            sid = <str>
            state_b64 = <str>
        }

        """
        sid = flask.request.args.get('sid', None)

        if sid is None:
            return make_response_json("No session id (sid) provided."), 400

        with self.controller:
            if not self.controller.has_session_uuid(sid):
                return make_response_json("session id '%s' not found" % sid,
                                          sid=sid), 404
            iqrs = self.controller.get_session(sid)
            iqrs.lock.acquire()  # lock BEFORE releasing controller

        try:
            iqrs_state_bytes = iqrs.get_state_bytes()
        finally:
            iqrs.lock.release()

        # Convert state bytes to a base64 string
        # - `b64encode` returns bytes, so decode to a string.
        iqrs_state_b64 = base64.b64encode(iqrs_state_bytes).decode('utf8')

        return make_response_json("Success",
                                  sid=sid,
                                  state_b64=iqrs_state_b64), 200

    # PUT /state
    def set_iqr_state(self):
        """ [See api.rst]
        Set the IQR session state for a given session ID.

        We expect the input bytes to have been generated by the matching
        get-state endpoint (see above). However, unlike the other endpoint's
        return format, the byte input to this endpoint must be encoded in
        URL-safe base64.

        Form Args:
            sid
                Session ID to set the input state to.
            state_base64
                Base64 of the state to set the session to.  This should be
                retrieved from the [GET] /state endpoint.

        Possible error code returns:
            400
                - No session ID provided.
                - No base64 bytes provided.
            404
                No session for the given ID.

        Success returns 200: {
            message = "Success"
            ...
            sid = <str>
        }

        """
        sid = flask.request.form.get('sid', None)
        state_base64 = flask.request.form.get('state_base64', None)

        if sid is None:
            return make_response_json("No session id (sid) provided."), 400
        elif state_base64 is None or len(state_base64) == 0:
            return make_response_json("No state package base64 provided."), 400

        # TODO: Limit the size of input state object? Is this already handled
        #       by other security measures?

        # Encoding is required because the b64decode does not handle being
        # given unicode (python2) or str (python3): needs bytes.

        try:
            # Using urlsafe version because it handles both regular and urlsafe
            # alphabets.
            state_bytes = \
                base64.urlsafe_b64decode(state_base64.encode('utf-8'))
        except (TypeError, binascii.Error) as ex:
            return make_response_json("Invalid base64 input: %s" % str(ex)), \
                   400

        with self.controller:
            if not self.controller.has_session_uuid(sid):
                return make_response_json("session id '%s' not found" % sid,
                                          sid=sid), 404
            iqrs = self.controller.get_session(sid)
            iqrs.lock.acquire()  # lock BEFORE releasing controller

        try:
            iqrs.set_state_bytes(state_bytes, self.descriptor_factory)
        finally:
            iqrs.lock.release()

        return make_response_json("Success", sid=sid), 200



"""
Descriptor computation helper utility. Checks data content type with respect
to the configured descriptor generator to skip content that does not match
the accepted types. Optionally, we can additionally filter out image content
whose image bytes we cannot load via ``PIL.Image.open``.
"""
import collections
import csv
import logging
import os

from smqtk.algorithms import get_descriptor_generator_impls
from smqtk.compute_functions import compute_many_descriptors
from smqtk.representation import (
    DescriptorElementFactory,
    get_data_set_impls,
    get_descriptor_index_impls,
)
from smqtk.representation.data_element.file_element import DataFileElement
from smqtk.utils.bin_utils import (
    utility_main_helper,
    report_progress,
    basic_cli_parser,
)
from smqtk.utils.image_utils import is_valid_element
from smqtk.utils import plugin, parallel


def default_config():
    return {
        "descriptor_generator":
            plugin.make_config(get_descriptor_generator_impls()),
        "descriptor_factory": DescriptorElementFactory.get_default_config(),
        "descriptor_index":
            plugin.make_config(get_descriptor_index_impls()),
        "optional_data_set":
            plugin.make_config(get_data_set_impls())
    }


def run_file_list(c, filelist_filepath, checkpoint_filepath, batch_size=None,
                  check_image=False):
    """
    Top level function handling configuration and inputs/outputs.

    :param c: Configuration dictionary (JSON)
    :type c: dict

    :param filelist_filepath: Path to a text file that lists paths to data
        files, separated by new lines.
    :type filelist_filepath: str

    :param checkpoint_filepath: Output file to which we write input filepath to
        SHA1 (UUID) relationships.
    :type checkpoint_filepath:

    :param batch_size: Optional batch size (None default) of data elements to
        process / descriptors to compute at a time. This causes files and
        stores to be written to incrementally during processing instead of
        one single batch transaction at a time.
    :type batch_size:

    :param check_image: Enable checking image loading from file before queueing
        that file for processing. If the check fails, the file is skipped
        instead of a halting exception being raised.
    :type check_image: bool

    """
    log = logging.getLogger(__name__)

    file_paths = [l.strip() for l in open(filelist_filepath)]

    log.info("Making descriptor factory")
    factory = DescriptorElementFactory.from_config(c['descriptor_factory'])

    log.info("Making descriptor index")
    #: :type: smqtk.representation.DescriptorIndex
    descriptor_index = plugin.from_plugin_config(c['descriptor_index'],
                                                 get_descriptor_index_impls())

    data_set = None
    if c['optional_data_set']['type'] is None:
        log.info("Not saving loaded data elements to data set")
    else:
        log.info("Initializing data set to append to")
        #: :type: smqtk.representation.DataSet
        data_set = plugin.from_plugin_config(c['optional_data_set'],
                                             get_data_set_impls())

    log.info("Making descriptor generator '%s'",
             c['descriptor_generator']['type'])
    #: :type: smqtk.algorithms.DescriptorGenerator
    generator = plugin.from_plugin_config(c['descriptor_generator'],
                                          get_descriptor_generator_impls())

    def iter_valid_elements():
        def is_valid(file_path):
            e = DataFileElement(file_path)

            if is_valid_element(
                    e, valid_content_types=generator.valid_content_types(),
                    check_image=check_image):
                return e
            else:
                return False

        data_elements = collections.deque()
        valid_files_filter = parallel.parallel_map(is_valid,
                                                   file_paths,
                                                   name="check-file-type",
                                                   use_multiprocessing=True)
        for dfe in valid_files_filter:
            if dfe:
                yield dfe
                if data_set is not None:
                    data_elements.append(dfe)
                    if batch_size and len(data_elements) == batch_size:
                        log.debug("Adding data element batch to set (size: %d)",
                                  len(data_elements))
                        data_set.add_data(*data_elements)
                        data_elements.clear()
        # elements only collected if we have a data-set configured, so add any
        # still in the deque to the set
        if data_elements:
            log.debug("Adding data elements to set (size: %d",
                      len(data_elements))
            data_set.add_data(*data_elements)

    log.info("Computing descriptors")
    m = compute_many_descriptors(iter_valid_elements(),
                                 generator,
                                 factory,
                                 descriptor_index,
                                 batch_size=batch_size,
                                 )

    # Recording computed file paths and associated file UUIDs (SHA1)
    cf = open(checkpoint_filepath, 'w')
    cf_writer = csv.writer(cf)
    try:
        rps = [0] * 7
        for de, descr in m:
            # We know that we are using DataFileElements going into the
            # compute_many_descriptors, so we can assume that's what comes out
            # of it as well.
            # noinspection PyProtectedMember
            cf_writer.writerow([de._filepath, descr.uuid()])
            report_progress(log.debug, rps, 1.)
    finally:
        del cf_writer
        cf.close()

    log.info("Done")


def cli_parser():
    parser = basic_cli_parser(__doc__)

    parser.add_argument('-b', '--batch-size',
                        type=int, default=0, metavar='INT',
                        help="Number of files to batch together into a single "
                             "compute async call. This defines the "
                             "granularity of the checkpoint file in regards "
                             "to computation completed. If given 0, we do not "
                             "batch and will perform a single "
                             "``compute_async`` call on the configured "
                             "generator. Default batch size is 0.")
    parser.add_argument('--check-image',
                        default=False, action='store_true',
                        help="If se should check image pixel loading before "
                             "queueing an input image for processing. If we "
                             "cannot load the image pixels via "
                             "``PIL.Image.open``, the input image is not "
                             "queued for processing")

    # Non-config required arguments
    g_required = parser.add_argument_group("Required Arguments")
    g_required.add_argument('-f', '--file-list',
                            type=str, default=None, metavar='PATH',
                            help="Path to a file that lists data file paths. "
                                 "Paths in this file may be relative, but "
                                 "will at some point be coerced into absolute "
                                 "paths based on the current working "
                                 "directory.")
    g_required.add_argument('-p', '--completed-files',
                            default=None, metavar='PATH',
                            help='Path to a file into which we add CSV '
                                 'format lines detailing filepaths that have '
                                 'been computed from the file-list provided, '
                                 'as the UUID for that data (currently the '
                                 'SHA1 checksum of the data).')

    return parser


def main():
    args = cli_parser().parse_args()
    config = utility_main_helper(default_config, args)
    log = logging.getLogger(__name__)

    completed_files_fp = args.completed_files
    filelist_fp = args.file_list
    batch_size = args.batch_size
    check_image = args.check_image

    # Input checking
    if not filelist_fp:
        log.error("No file-list file specified")
        exit(102)
    elif not os.path.isfile(filelist_fp):
        log.error("Invalid file list path: %s", filelist_fp)
        exit(103)

    if not completed_files_fp:
        log.error("No complete files output specified")
        exit(104)

    if batch_size < 0:
        log.error("Batch size must be >= 0.")
        exit(105)

    run_file_list(
        config,
        filelist_fp,
        completed_files_fp,
        batch_size,
        check_image
    )


if __name__ == '__main__':
    main()



import unittest

import mock

from smqtk.representation.descriptor_element.solr_element import \
    SolrDescriptorElement


if SolrDescriptorElement.is_usable():

    class TestDescriptorSolrElement (unittest.TestCase):

        TEST_URL = 'http://localhost:8983/solr'

        @mock.patch("solr.Solr")
        def test_configuration(self, _mock_Solr):
            default_config = SolrDescriptorElement.get_default_config()
            self.assertEqual(default_config, {
                "solr_conn_addr": None,
                "type_field": None,
                "uuid_field": None,
                "vector_field": None,
                "timestamp_field": None,
                "timeout": 10,
                "persistent_connection": False,
                "commit_on_set": True,
            })

            default_config.update({
                'solr_conn_addr': self.TEST_URL,
                "type_field": "type_s",
                "uuid_field": "uuid_s",
                "vector_field": "descriptor_fs",
                "timestamp_field": "timestamp_f",
                "timeout": 101,
                "persistent_connection": True,
                "commit_on_set": False,
            })

            # Result instance should have parameters matching those in config.
            #: :type: SolrDescriptorElement
            inst1 = SolrDescriptorElement.from_config(default_config,
                                                      'test', 'a')
            self.assertEqual(default_config, inst1.get_config())
            self.assertEqual(inst1.solr_conn_addr, self.TEST_URL)
            self.assertEqual(inst1.solr_timeout, 101)
            self.assertEqual(inst1.solr_persistent_connection, True)
            self.assertEqual(inst1.type_field, 'type_s')
            self.assertEqual(inst1.uuid_field, 'uuid_s')
            self.assertEqual(inst1.vector_field, 'descriptor_fs')
            self.assertEqual(inst1.timestamp_field, 'timestamp_f')
            self.assertEqual(inst1.solr_commit_on_set, False)

            # State parameters should be equal.
            #: :type: SolrDescriptorElement
            inst2 = SolrDescriptorElement.from_config(inst1.get_config(),
                                                      'test', 'a')
            self.assertEqual(inst1.type_field, inst2.type_field)
            self.assertEqual(inst1.uuid_field, inst2.uuid_field)
            self.assertEqual(inst1.vector_field, inst2.vector_field)
            self.assertEqual(inst1.timestamp_field, inst2.timestamp_field)
            self.assertEqual(inst1.solr_conn_addr, inst2.solr_conn_addr)
            self.assertEqual(inst1.solr_timeout, inst2.solr_timeout)
            self.assertEqual(inst1.solr_persistent_connection,
                             inst2.solr_persistent_connection)
            self.assertEqual(inst1.solr_commit_on_set, inst2.solr_commit_on_set)



"""
Utility for transforming an input image in various standardized ways, saving
out those transformed images with standard namings. Transformations used are
configurable via a configuration file (JSON).

Configuration details:
{
    "crop": {

        "center_levels": null | int
            # If greater than 0, crop out one or more increasing smaller images
            # from a base image by cutting off increasingly larger portions of
            # the outside perimeter. Cropped image dimensions determined by the
            # dimensions of the base image and the number of crops to generate.

        "quadrant_pyramid_levels": null | int
            # If greater than 0, generate a number of crops based on a number of
            # quad-tree partitions made based on the given number of levels.
            # Partitions for all levels less than the level provides are also
            # made.

        "tile_shape": null | [width, height]
            # If not null and is a list of two integers, crop out tile windows
            # from the base image that have the width and height specified.
            # If the image width or height is not evenly divisible by the tile
            # width or height, respectively, then the crop out as many tiles as
            # neatly fit starting from the axis origin. The remaining pixels are
            # ignored.

        "tile_stride": null | [x, y]
            # If not null and is a list of two integers, crop out sub-images of
            # the above width and height (if given) with this stride. When not
            # this is not provided, the default stride is the same as the tile
            # width and height.
    },

    "brightness_levels": null | int
        # Generate a number of images with different brightness levels using
        # linear interpolation to choose levels between 0 (black) and 1
        # (original image) as well as between 1 and 2.
        # Results will not include contrast level 0, 1 or 2 images.

    "contrast_levels": null | int
        # Generate a number of images with different contrast levels using
        # linear interpolation to choose levels between 0 (black) and 1
        # (original image) as well as between 1 and 2.
        # Results will not include contrast level 0, 1 or 2 images.

}

"""

import logging
import os

import PIL.Image
import PIL.ImageEnhance
import numpy

import smqtk.utils.bin_utils
import smqtk.utils.file_utils
import smqtk.utils.parallel

from six.moves import range, zip


__author__ = "paul.tunison@kitware.com"


def image_crop_center_levels(image, n_crops):
    """
    Crop out one or more increasing smaller images from a base image by cutting
    off increasingly larger portions of the outside perimeter. Cropped image
    dimensions determined by the dimensions of the base image and the number of
    crops to generate.

    :param image: The base image array. This is not modified,.
    :type image: PIL.Image.Image

    :param n_crops: Number of crops to generate.
    :type n_crops: int

    :return: Generator yielding paired level number and PIL.Image.Image tuples.
        Cropped images have not loaded/copied yet, so changes to the original
        image will affect them.
    :rtype: __generator[(int, PIL.Image.Image)]

    """
    n_crops = int(n_crops)
    if n_crops <= 0:
        raise ValueError("Can't produce 0 or negative crops")

    #: :type: numpy.ndarray
    x_points = numpy.linspace(0, image.width, 2 + n_crops * 2, dtype=int)
    #: :type: numpy.ndarray
    y_points = numpy.linspace(0, image.height, 2 + n_crops * 2, dtype=int)

    # Outside edges of generated points in the original image size
    for i in range(1, n_crops + 1):
        # crop wants: [xmin, ymin, xmax, ymax]
        t = list(zip(x_points[[i, -i - 1]], y_points[[i, -i - 1]]))
        yield i, image.crop(t[0] + t[1])


def image_crop_quadrant_pyramid(image, n_levels):
    """
    Generate a number of crops based on a number of quadrant sub-partitions made
    based on the given number of levels.

    For example, 1 level would yield 4 crops from the 2x2 partition of the
    image. 2 levels would yield 20 crops for the 2x2 partition and the 4x4
    partition. 3 levels would yield the partitions of 2x2, 4x4 and 8x8 yielding
    84 crops, etc. General rule: (2^i)^2 for i in [1 n_levels]

    :param image: Image to crop in quadrant partitions
    :type image: PIL.Image.Image

    :param n_levels: Number of quadrant levels to generate crops for.
    :type n_levels: int

    :return: Generator yielding paired level, quadrant position and
        PIL.Image.Image tuples. Quadrant position is in (x, y) format.
        Crop images have not loaded/copied yet, so changes to the original
        image will affect them.
    :rtype: __generator[(int, (int, int), PIL.Image.Image)]

    """
    n_crops = int(n_levels)
    if n_crops <= 0:
        raise ValueError("Can't produce 0 or negative levels")

    for l in range(1, n_levels + 1):
        l_sq = 2**l
        xs = numpy.linspace(0, image.width, l_sq + 1, endpoint=True, dtype=int)
        ys = numpy.linspace(0, image.height, l_sq + 1, endpoint=True, dtype=int)
        for j in range(l_sq):
            for i in range(l_sq):
                yield (
                    l,
                    (i, j),
                    image.crop([xs[i], ys[j], xs[i + 1], ys[j + 1]])
                )


def image_crop_tiles(image, tile_width, tile_height, stride=None):
    """
    Crop out tile windows from the base image that have the width and height
    specified.

    If the image width or height is not evenly divisible by the tile width or
    height, respectively, then the crop out as many tiles as neatly fit starting
    from the axis origin. The remaining pixels are ignored.

    :param image: Image to crop tiles from.
    :type image: PIL.Image.Image

    :param tile_width: Tile crop width in pixels.
    :type tile_width: int

    :param tile_height: Tile crop height in pixels.
    :type tile_height: int

    :param stride: Optional tuple of integer pixel stride for cropping out sub-
        images. When this is None, the stride is the same as the width and
        height of the requested sub-images.
    :type stride: None | (int, int)

    :return: Generator yielding tuples containing a cropped image and its upper-
        left xy position in the original image.
    :rtype: __generator[(int, int, PIL.Image.Image)]

    """
    if stride:
        stride_x, stride_y = map(int, stride)
    else:
        stride_x = tile_width
        stride_y = tile_height

    # upper-left xy pixel coordinates for sub-images.
    y = 0
    while (y + tile_height) < image.height:
        x = 0
        while (x + tile_width) < image.width:
            t = image.crop([x, y, x+tile_width, y+tile_height])
            yield (x, y, t)
            x += stride_x
        y += stride_y


def image_brightness_intervals(image, n):
    """
    Generate a number of images with different brightness levels using linear
    interpolation to choose levels between 0 (black) and 1 (original image) as
    well as between 1 and 2.

    Results will not include contrast level 0, 1 or 2 images.

    """
    n = int(n)
    if n <= 0:
        raise ValueError("Can't produce 0 intervals")

    b = numpy.linspace(0, 1, n+2, endpoint=True, dtype=float)
    for v in b[1:-1]:
        yield v, PIL.ImageEnhance.Brightness(image).enhance(v)
    b = numpy.linspace(1, 2, n + 2, endpoint=True, dtype=float)
    for v in b[1:-1]:
        yield v, PIL.ImageEnhance.Brightness(image).enhance(v)


def image_contrast_intervals(image, n):
    """
    Generate a number of images with different contrast levels using linear
    interpolation to choose levels between 0 (black) and 1 (original image) as
    well as between 1 and 2.

    Results will not include contrast level 0, 1 or 2 images.

    """
    n = int(n)
    if n <= 0:
        raise ValueError("Can't produce 0 intervals")

    b = numpy.linspace(0, 1, n + 2, endpoint=True, dtype=float)
    for v in b[1:-1]:
        yield v, PIL.ImageEnhance.Contrast(image).enhance(v)
    b = numpy.linspace(1, 2, n + 2, endpoint=True, dtype=float)
    for v in b[1:-1]:
        yield v, PIL.ImageEnhance.Contrast(image).enhance(v)


def generate_image_transformations(image_path,
                                   crop_center_n, crop_quadrant_levels,
                                   crop_tile_shape, crop_tile_stride,
                                   brightness_intervals,
                                   contrast_intervals,
                                   output_dir=None,
                                   output_ext='.png'):
    """
    Transform an input image into different crops or other transforms,
    outputting results to the given output directory without overwriting or
    otherwise changing the input image.

    By default, if not told otherwise, we will write output images in the same
    directory as the source image. Output images share a core filename as the
    source image, but with extra suffix syntax to differentiate produced images
    from the original. Output images will share the same image extension as the
    source image.
    """
    log = logging.getLogger(__name__)

    abs_path = os.path.abspath(image_path)
    output_dir = output_dir or os.path.dirname(abs_path)
    smqtk.utils.file_utils.safe_create_dir(output_dir)
    p_base = os.path.splitext(os.path.basename(abs_path))[0]
    p_ext = output_ext
    p_base = os.path.join(output_dir, p_base)
    image = PIL.Image.open(image_path).convert('RGB')

    def save_image(img, suffixes):
        """
        Save an image based on source image basename and an iterable of suffix
        parts that will be separated by periods.
        """
        fn = '.'.join([p_base] + list(suffixes)) + p_ext
        log.debug("Saving: %s", fn)
        img.save(fn)

    if crop_center_n:
        log.info("Computing center crops")
        tag = "crop_centers"
        for l, c in image_crop_center_levels(image, crop_center_n):
            save_image(c, [tag, str(l)])

    if crop_quadrant_levels:
        log.info("Computing quadrant crops")
        tag = "crop_quadrants"
        for l, (i, j), c in image_crop_quadrant_pyramid(image,
                                                        crop_quadrant_levels):
            save_image(c, [tag, str(l), "q_{:d}_{:d}".format(i, j)])

    if crop_tile_shape and crop_tile_shape[0] > 0 and crop_tile_shape[1] > 0:
        tag = "crop_tiles"
        t_width = crop_tile_shape[0]
        t_height = crop_tile_shape[1]
        log.info("Cropping %dx%d pixel tiles from images with stride %s"
                 % (t_width, t_height, crop_tile_stride))
        # List needed to iterate generator.
        list(smqtk.utils.parallel.parallel_map(
            lambda x, y, ii:
                save_image(ii, [tag,
                                '%dx%d+%d+%d' % (t_width, t_height, x, y)]),
            image_crop_tiles(image, t_width, t_height, crop_tile_stride)
        ))

    if brightness_intervals:
        log.info("Computing brightness variants")
        for b, i in image_brightness_intervals(image, brightness_intervals):
            save_image(i, ['brightness', str(b)])

    if contrast_intervals:
        log.info("Computing contrast variants")
        for c, i in image_contrast_intervals(image, contrast_intervals):
            save_image(i, ['contrast', str(c)])


def default_config():
    return {
        "crop": {
            # 0 means disabled
            "center_levels": None,
            # 0 means disabled, 2 meaning 2x2 and 4x4
            "quadrant_pyramid_levels": None,
            # Tile shape or None for no tiling
            "tile_shape": None,
            # The stride of tiles top crop out. This defaults to the height and
            # width of the tiles to create non-overlapping chips.
            "tile_stride": None,
        },
        "brightness_levels": None,
        "contrast_levels": None,
    }


def cli_parser():
    parser = smqtk.utils.bin_utils.basic_cli_parser(__doc__)

    g_io = parser.add_argument_group("Input/Output")
    g_io.add_argument("-i", "--image",
                      help="Image to produce transformations for.")
    g_io.add_argument("-o", "--output",
                      help="Directory to output generated images to. By "
                           "default, if not told otherwise, we will write "
                           "output images in the same directory as the source "
                           "image. Output images share a core filename as the "
                           "source image, but with extra suffix syntax to "
                           "differentiate produced images from the original. "
                           "Output images will share the same image extension "
                           "as the source image.")
    return parser


def main():
    args = cli_parser().parse_args()
    config = smqtk.utils.bin_utils.utility_main_helper(default_config, args)
    input_image_path = args.image
    output_dir = args.output

    if input_image_path is None:
        raise ValueError("No input image path given")

    crop_center_levels = config['crop']['center_levels']
    crop_quad_levels = config['crop']['quadrant_pyramid_levels']
    crop_tile_shape = config['crop']['tile_shape']
    crop_tile_stride = config['crop']['tile_stride']
    b_levels = config['brightness_levels']
    c_levels = config['contrast_levels']

    generate_image_transformations(
        input_image_path,
        crop_center_levels, crop_quad_levels,
        crop_tile_shape, crop_tile_stride,
        b_levels, c_levels,
        output_dir
    )


if __name__ == '__main__':
    main()



from __future__ import division, print_function
import inspect
import os
import pickle
import unittest

import PIL.Image
import mock
import numpy

from smqtk.algorithms.descriptor_generator import get_descriptor_generator_impls
from smqtk.algorithms.descriptor_generator.caffe_descriptor import \
    caffe, CaffeDescriptorGenerator
# Testing protected helper function
# noinspection PyProtectedMember
from smqtk.algorithms.descriptor_generator.caffe_descriptor import \
    _process_load_img_array
from smqtk.representation.data_element import from_uri
from smqtk.representation.data_element.url_element import DataUrlElement
from smqtk.tests import TEST_DATA_DIR


if CaffeDescriptorGenerator.is_usable():

    class TestCaffeDesctriptorGenerator (unittest.TestCase):

        lenna_image_fp = os.path.join(TEST_DATA_DIR, 'Lenna.png')
        lenna_alexnet_fc7_descr_fp = \
            os.path.join(TEST_DATA_DIR, 'Lenna.alexnet_fc7_output.npy')

        # Dummy Caffe configuration files + weights
        # - weights is actually an empty file (0 bytes), which caffe treats
        #   as random/zero values (not sure exactly what's happening, but
        #   always results in a zero-vector).
        dummy_net_topo_fp = \
            os.path.join(TEST_DATA_DIR, 'caffe.dummpy_network.prototxt')
        dummy_caffe_model_fp = \
            os.path.join(TEST_DATA_DIR, 'caffe.empty_model.caffemodel')
        dummy_img_mean_fp = \
            os.path.join(TEST_DATA_DIR, 'caffe.dummy_mean.npy')

        www_uri_alexnet_prototxt = \
            'https://data.kitware.com/api/v1/file/57e2f3fd8d777f10f26e532c' \
            '/download'
        www_uri_alexnet_caffemodel = \
            'https://data.kitware.com/api/v1/file/57dae22f8d777f10f26a2a86' \
            '/download'
        www_uri_image_mean_proto = \
            'https://data.kitware.com/api/v1/file/57dae0a88d777f10f26a2a82' \
            '/download'

        def test_impl_findable(self):
            self.assertIn(CaffeDescriptorGenerator.__name__,
                          get_descriptor_generator_impls())

        @mock.patch('smqtk.algorithms.descriptor_generator.caffe_descriptor'
                    '.CaffeDescriptorGenerator._setup_network')
        def test_get_config(self, _m_cdg_setupNetwork):
            # Mocking set_network so we don't have to worry about actually
            # initializing any caffe things for this test.
            expected_params = {
                'network_prototxt_uri': 'some_prototxt_uri',
                'network_model_uri': 'some_caffemodel_uri',
                'image_mean_uri': 'some_imagemean_uri',
                'return_layer': 'layer name',
                'batch_size': 777,
                'use_gpu': False,
                'gpu_device_id': 8,
                'network_is_bgr': False,
                'data_layer': 'data-other',
                'load_truncated_images': True,
                'pixel_rescale': (.2, .8),
                'input_scale': 1.5,
            }
            # make sure that we're considering all constructor parameter options
            expected_param_keys = \
                set(inspect.getargspec(CaffeDescriptorGenerator.__init__)
                           .args[1:])
            self.assertSetEqual(set(expected_params.keys()),
                                expected_param_keys)
            g = CaffeDescriptorGenerator(**expected_params)
            self.assertEqual(g.get_config(), expected_params)

        @mock.patch('smqtk.algorithms.descriptor_generator.caffe_descriptor'
                    '.CaffeDescriptorGenerator._setup_network')
        def test_pickle_save_restore(self, m_cdg_setupNetwork):
            # Mocking set_network so we don't have to worry about actually
            # initializing any caffe things for this test.
            expected_params = {
                'network_prototxt_uri': 'some_prototxt_uri',
                'network_model_uri': 'some_caffemodel_uri',
                'image_mean_uri': 'some_imagemean_uri',
                'return_layer': 'layer name',
                'batch_size': 777,
                'use_gpu': False,
                'gpu_device_id': 8,
                'network_is_bgr': False,
                'data_layer': 'data-other',
                'load_truncated_images': True,
                'pixel_rescale': (.2, .8),
                'input_scale': 1.5,
            }
            g = CaffeDescriptorGenerator(**expected_params)
            # Initialization sets up the network on construction.
            self.assertEqual(m_cdg_setupNetwork.call_count, 1)

            g_pickled = pickle.dumps(g, -1)
            g2 = pickle.loads(g_pickled)
            # Network should be setup for second class class just like in
            # initial construction.
            self.assertEqual(m_cdg_setupNetwork.call_count, 2)

            self.assertIsInstance(g2, CaffeDescriptorGenerator)
            self.assertEqual(g.get_config(), g2.get_config())

        @mock.patch('smqtk.algorithms.descriptor_generator.caffe_descriptor'
                    '.CaffeDescriptorGenerator._setup_network')
        def test_invalid_datatype(self, _m_cdg_setupNetwork):
            # Test that a data element with an incorrect content type raises an
            # exception.

            # Passing purposefully bag constructor parameters and ignoring
            # Caffe network setup (above mocking).
            # noinspection PyTypeChecker
            g = CaffeDescriptorGenerator(None, None, None)
            bad_element = from_uri(os.path.join(TEST_DATA_DIR, 'test_file.dat'))
            self.assertRaises(
                ValueError,
                g.compute_descriptor,
                bad_element
            )

        def test_process_load_img(self):
            # using image shape, meaning no transformation should occur
            test_data_layer = 'data'
            test_transformer = \
                caffe.io.Transformer({test_data_layer: (1, 3, 512, 512)})

            lenna_elem = from_uri(self.lenna_image_fp)
            a_expected = numpy.asarray(PIL.Image.open(self.lenna_image_fp),
                                       numpy.float32)
            a = _process_load_img_array((
                lenna_elem, test_transformer, test_data_layer, None, None
            ))
            numpy.testing.assert_allclose(a, a_expected)

        @mock.patch('smqtk.algorithms.descriptor_generator.caffe_descriptor'
                    '.CaffeDescriptorGenerator._setup_network')
        def test_no_internal_compute_descriptor(self, _m_cdg_setupNetwork):
            # This implementation's descriptor computation logic sits in async
            # method override due to caffe's natural multi-element computation
            # interface. Thus, ``_compute_descriptor`` should not be
            # implemented.

            # Passing purposefully bag constructor parameters and ignoring
            # Caffe network setup (above mocking).
            # noinspection PyTypeChecker
            g = CaffeDescriptorGenerator(0, 0, 0)
            self.assertRaises(
                NotImplementedError,
                g._compute_descriptor, None
            )

        def test_compute_descriptor_dummy_model(self):
            # Caffe dummy network interaction test Lenna image)

            # Construct network with an empty model just to see that our
            # interaction with the Caffe API is successful. We expect a
            # zero-valued descriptor vector.
            g = CaffeDescriptorGenerator(self.dummy_net_topo_fp,
                                         self.dummy_caffe_model_fp,
                                         self.dummy_img_mean_fp,
                                         return_layer='fc', use_gpu=False)
            d = g.compute_descriptor(from_uri(self.lenna_image_fp))
            self.assertAlmostEqual(d.vector().sum(), 0., 12)

        @unittest.skipUnless(DataUrlElement.is_usable(),
                             "URL resolution not functional")
        def test_compute_descriptor_from_url_lenna_description(self):
            # Caffe AlexNet interaction test (Lenna image)
            # This is a long test since it has to download data for remote URIs
            d = CaffeDescriptorGenerator(
                self.www_uri_alexnet_prototxt,
                self.www_uri_alexnet_caffemodel,
                self.www_uri_image_mean_proto,
                return_layer='fc7',
                use_gpu=False,
            )
            lenna_elem = from_uri(self.lenna_image_fp)
            expected_descr = numpy.load(self.lenna_alexnet_fc7_descr_fp)
            descr = d.compute_descriptor(lenna_elem).vector()
            numpy.testing.assert_allclose(descr, expected_descr, atol=1e-5)

        def test_compute_descriptor_async_no_data(self):
            # Should get a ValueError when given no descriptors to async method
            g = CaffeDescriptorGenerator(self.dummy_net_topo_fp,
                                         self.dummy_caffe_model_fp,
                                         self.dummy_img_mean_fp,
                                         return_layer='fc', use_gpu=False)
            self.assertRaises(
                ValueError,
                g.compute_descriptor_async, []
            )



from ._interface_classifier import Classifier
from ._interface_supervised import SupervisedClassifier
from ._get_impls import get_classifier_impls

# Requires ``get_classifier_impls`` function.
from ._classifier_collection import ClassifierCollection



"""
This module contains a general locality-sensitive-hashing algorithm for
nearest neighbor indexing, and various implementations of LSH functors for use
in the base.
"""
import collections
import itertools
import multiprocessing

import numpy
from six.moves import map, zip

from smqtk.algorithms.nn_index import NearestNeighborsIndex
from smqtk.algorithms.nn_index.hash_index import get_hash_index_impls
from smqtk.algorithms.nn_index.hash_index.linear import LinearHashIndex
from smqtk.algorithms.nn_index.lsh.functors import get_lsh_functor_impls
from smqtk.exceptions import ReadOnlyError
from smqtk.representation import get_descriptor_index_impls, \
    get_key_value_store_impls
from smqtk.representation.descriptor_element import elements_to_matrix
from smqtk.utils import metrics
from smqtk.utils import plugin
from smqtk.utils.bit_utils import bit_vector_to_int_large
from smqtk.utils.bin_utils import ProgressReporter
from smqtk.utils import merge_dict


class LSHNearestNeighborIndex (NearestNeighborsIndex):
    """
    Locality-sensitive hashing based nearest neighbor index

    This type of algorithm relies on a hashing algorithm to hash descriptors
    such that similar descriptors are hashed the same or similar hash value.
    This allows simpler distance functions (hamming distance) to be performed on
    hashes in order to find nearby bins which are more likely to hold similar
    descriptors.

    LSH nearest neighbor algorithms consist of:
        * Index of descriptors to query over
        * A hashing function that transforms a descriptor vector into a
          hash (bit-vector).
        * Key-Value store of hash values to their set of hashed descriptor
          UUIDs.
        * Nearest neighbor index for indexing bit-vectors (treated as
          descriptors)

    """

    @classmethod
    def is_usable(cls):
        # This "shell" class is always usable, no special dependencies.
        return True

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        By default, we observe what this class's constructor takes as
        arguments, turning those argument names into configuration dictionary
        keys. If any of those arguments have defaults, we will add those values
        into the configuration dictionary appropriately. The dictionary
        returned should only contain JSON compliant value types.

        It is not be guaranteed that the configuration dictionary returned
        from this method is valid for construction of an instance of this
        class.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        """
        default = super(LSHNearestNeighborIndex, cls).get_default_config()

        lf_default = plugin.make_config(get_lsh_functor_impls())
        default['lsh_functor'] = lf_default

        di_default = plugin.make_config(get_descriptor_index_impls())
        default['descriptor_index'] = di_default

        hi_default = plugin.make_config(get_hash_index_impls())
        default['hash_index'] = hi_default
        default['hash_index_comment'] = "'hash_index' may also be null to " \
                                        "default to a linear index built at " \
                                        "query time."

        h2u_default = plugin.make_config(get_key_value_store_impls())
        default['hash2uuids_kvstore'] = h2u_default

        return default

    @classmethod
    def from_config(cls, config_dict, merge_default=True):
        """
        Instantiate a new instance of this class given the configuration
        JSON-compliant dictionary encapsulating initialization arguments.

        This method should not be called via super unless and instance of the
        class is desired.

        :param config_dict: JSON compliant dictionary encapsulating
            a configuration.
        :type config_dict: dict

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: LSHNearestNeighborIndex

        """
        # Controlling merge here so we can control known comment stripping from
        # default config.
        if merge_default:
            merged = cls.get_default_config()
            merge_dict(merged, config_dict)
        else:
            merged = config_dict

        merged['lsh_functor'] = \
            plugin.from_plugin_config(merged['lsh_functor'],
                                      get_lsh_functor_impls())
        merged['descriptor_index'] = \
            plugin.from_plugin_config(merged['descriptor_index'],
                                      get_descriptor_index_impls())

        # Hash index may be None for a default at-query-time linear indexing
        if merged['hash_index'] and merged['hash_index']['type']:
            merged['hash_index'] = \
                plugin.from_plugin_config(merged['hash_index'],
                                          get_hash_index_impls())
        else:
            cls.get_logger().debug("No HashIndex impl given. Passing ``None``.")
            merged['hash_index'] = None

        # remove possible comment added by default generator
        if 'hash_index_comment' in merged:
            del merged['hash_index_comment']

        merged['hash2uuids_kvstore'] = \
            plugin.from_plugin_config(merged['hash2uuids_kvstore'],
                                      get_key_value_store_impls())

        return super(LSHNearestNeighborIndex, cls).from_config(merged, False)

    def __init__(self, lsh_functor, descriptor_index, hash2uuids_kvstore,
                 hash_index=None,
                 distance_method='cosine', read_only=False):
        """
        Initialize LSH algorithm with a hashing functor, descriptor index and
        hash nearest-neighbor index.

        In order to provide out-of-the-box neighbor querying ability, at least
        the ``descriptor_index`` and ``hash2uuids_kvstore`` must be provided.
        The UIDs of descriptors in the ``descriptor_index`` should be fully
        mapped by the key-value mapping (``hash2uuids_kvstore``). If not, not
        all descriptors will be accessible via the nearest-neighbor query (not
        referenced in ``hash2uuids_kvstore`` map), or the requested number of
        neighbors might not be returned (descriptors hashed in ``hash_index``
        disjoint from ``descriptor_index``).

        An ``LSHNearestNeighborIndex`` instance is effectively read-only if any
        of its input structures (`descriptor_index`, `hash2uuids_kvstore`,
        `hash_index`) are read-only.

        :param lsh_functor: LSH functor implementation instance.
        :type lsh_functor: smqtk.algorithms.nn_index.lsh.functors.LshFunctor

        :param descriptor_index: Index in which DescriptorElements will be
            stored.
        :type descriptor_index: smqtk.representation.DescriptorIndex

        :param hash2uuids_kvstore: KeyValueStore instance to use for linking a
            hash code, as an integer, in the ``hash_index`` with one or more
            ``DescriptorElement`` instance UUIDs in the given
            ``descriptor_index``.
        :type hash2uuids_kvstore: smqtk.representation.KeyValueStore

        :param hash_index: ``HashIndex`` for indexing unique hash codes using
            hamming distance.

            If this is set to ``None`` (default), we will perform brute-force
            linear neighbor search for each query based on the hash codes
            currently in the hash2uuid index using hamming distance
        :type hash_index: smqtk.algorithms.nn_index.hash_index.HashIndex | None

        :param distance_method: String label of distance method to use for
            determining descriptor similarity (after finding near hashes for a
            given query).

            This must one of the following:
                - "euclidean": Simple euclidean distance between two
                    descriptors (L2 norm).
                - "cosine": Cosine angle distance/similarity between two
                    descriptors.
                - "hik": Histogram intersection distance between two
                    descriptors.
        :type distance_method: str

        :param read_only: If this index should only read from its configured
            descriptor and hash indexes. This will cause a ``ReadOnlyError`` to
            be raised from build_index.
        :type read_only: bool

        :raises ValueError: Invalid distance method specified.

        """
        super(LSHNearestNeighborIndex, self).__init__()

        # TODO(paul.tunison): Add in-memory empty defaults for
        #   descriptor_index/hash2uuids_kvstore attributes.
        self.lsh_functor = lsh_functor
        self.descriptor_index = descriptor_index
        self.hash_index = hash_index
        # Will use with int|long keys and set[collection.Hashable] values.
        self.hash2uuids_kvstore = hash2uuids_kvstore
        self.distance_method = distance_method
        self.read_only = read_only

        # Lock for model component access (combination of descriptor-set,
        # hash_index and kvstore).  Multiprocessing because resources can be
        # potentially modified on other processes.
        self._model_lock = multiprocessing.RLock()

        self._distance_function = self._get_dist_func(self.distance_method)

    @staticmethod
    def _get_dist_func(distance_method):
        """
        Return appropriate distance function given a string label
        """
        if distance_method == "euclidean":
            return metrics.euclidean_distance
        elif distance_method == "cosine":
            # Inverse of cosine similarity function return
            return metrics.cosine_distance
        elif distance_method == 'hik':
            return metrics.histogram_intersection_distance_fast
        else:
            # TODO: Support scipy/scikit-learn distance methods
            raise ValueError("Invalid distance method label. Must be one of "
                             "['euclidean' | 'cosine' | 'hik']")

    def get_config(self):
        hi_conf = None
        if self.hash_index is not None:
            hi_conf = plugin.to_plugin_config(self.hash_index)
        return {
            "lsh_functor": plugin.to_plugin_config(self.lsh_functor),
            "descriptor_index": plugin.to_plugin_config(self.descriptor_index),
            "hash_index": hi_conf,
            "hash2uuids_kvstore":
                plugin.to_plugin_config(self.hash2uuids_kvstore),
            "distance_method": self.distance_method,
            "read_only": self.read_only,
        }

    def count(self):
        """
        :return: Maximum number of descriptors reference-able via a
            nearest-neighbor query (count of descriptor index). Actual return
            may be smaller of hash2uuids mapping is not complete.
        :rtype: int
        """
        with self._model_lock:
            c = 0
            for set_v in self.hash2uuids_kvstore.values():
                c += len(set_v)
            return c

    def _build_index(self, descriptors):
        """
        Internal method to be implemented by sub-classes to build the index with
        the given descriptor data elements.

        Subsequent calls to this method should rebuild the current index.  This
        method shall not add to the existing index nor raise an exception to as
        to protect the current index.

        :raises ReadOnlyError: This index is set to be read-only and cannot be
            modified.

        :param descriptors: Iterable of descriptor elements to build index
            over.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """
        with self._model_lock:
            if self.read_only:
                raise ReadOnlyError("Cannot modify container attributes due to "
                                    "being in read-only mode.")

            self._log.debug("Clearing and adding new descriptor elements")
            self.descriptor_index.clear()
            self.descriptor_index.add_many_descriptors(descriptors)

            self._log.debug("Generating hash codes")
            #: :type: collections.deque[numpy.ndarray[bool]]
            hash_vectors = collections.deque()
            self.hash2uuids_kvstore.clear()
            prog_reporter = ProgressReporter(self._log.debug, 1.0).start()
            # We just cleared the previous store, so aggregate new kv-mapping
            # in ``kvstore_update`` for single update after loop.
            kvstore_update = collections.defaultdict(set)
            for d in self.descriptor_index:
                h_vec = self.lsh_functor.get_hash(d.vector())
                hash_vectors.append(h_vec)
                h_int = bit_vector_to_int_large(h_vec)
                kvstore_update[h_int] |= {d.uuid()}
                prog_reporter.increment_report()
            prog_reporter.report()
            self.hash2uuids_kvstore.add_many(kvstore_update)
            del kvstore_update

            if self.hash_index is not None:
                self._log.debug("Clearing and building hash index of type %s",
                                type(self.hash_index))
                # a build is supposed to clear previous state.
                self.hash_index.build_index(hash_vectors)

    def _update_index(self, descriptors):
        """
        Internal method to be implemented by sub-classes to additively update
        the current index with the one or more descriptor elements given.

        If no index exists yet, a new one should be created using the given
        descriptors.

        :raises ReadOnlyError: This index is set to be read-only and cannot be
            modified.

        :param descriptors: Iterable of descriptor elements to add to this
            index.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """
        with self._model_lock:
            if self.read_only:
                raise ReadOnlyError("Cannot modify container attributes due "
                                    "to being in read-only mode.")
            # tee out iterable for use in adding to index as well as hash code
            # generation.
            d_for_index, d_for_hashing = itertools.tee(descriptors, 2)

            self._log.debug("Updating descriptor index.")
            self.descriptor_index.add_many_descriptors(d_for_index)

            self._log.debug("Generating hash codes for new descriptors")
            prog_reporter = ProgressReporter(self._log.debug, 1.0).start()
            #: :type: collections.deque[numpy.ndarray[bool]]
            hash_vectors = collections.deque()  # for updating hash_index
            # for updating kv-store after collecting new hash codes
            kvstore_update = {}
            for d in d_for_hashing:
                h_vec = self.lsh_functor.get_hash(d.vector())
                hash_vectors.append(h_vec)
                h_int = bit_vector_to_int_large(h_vec)
                # Get, update and reinsert hash UUID set object.
                if h_int not in kvstore_update:
                    #: :type: set
                    kvstore_update[h_int] = \
                        self.hash2uuids_kvstore.get(h_int, set())
                kvstore_update[h_int] |= {d.uuid()}
                prog_reporter.increment_report()
            prog_reporter.report()

            self._log.debug("Updating kv-store with new hash codes")
            self.hash2uuids_kvstore.add_many(kvstore_update)
            del kvstore_update

            if self.hash_index is not None:
                self._log.debug("Updating hash index structure.")
                self.hash_index.update_index(hash_vectors)

    def _remove_from_index(self, uids):
        """
        Remove descriptors from this index associated with the given UIDs.

        :param uids: Iterable of UIDs of descriptors to remove from this index.
        :type uids: collections.Iterable[collections.Hashable]

        :raises KeyError: One or more UIDs provided do not match any stored
            descriptors.  The index should not be modified.
        :raises ReadOnlyError: This index is set to be read-only and cannot be
            modified.

        """
        with self._model_lock:
            if self.read_only:
                raise ReadOnlyError("Cannot modify container attributes due "
                                    "to being in read-only mode.")

            uids = list(uids)

            # Remove UIDs from our hash2uid-kvs
            # - get the hash for each input UID's descriptor, remove UID from
            #   recorded association set.
            # - `get_many_descriptors` fails when bad UIDs are provided
            #   (KeyError).
            self._log.debug("Removing hash2uid entries for UID's descriptors")
            h_vectors = collections.deque()
            h_ints = collections.deque()
            for d in self.descriptor_index.get_many_descriptors(uids):
                h_vec = self.lsh_functor.get_hash(d.vector())
                h_vectors.append(h_vec)
                h_int = bit_vector_to_int_large(h_vec)
                h_ints.append(h_int)

            # If we're here, then all given UIDs mapped to an indexed
            # descriptor.  Proceed with removal from hash2uids kvs.  If a hash
            # no longer maps anything, remove that key from the KVS.
            hashes_for_removal = collections.deque()
            # store key-value pairs to update after loop in batch call
            kvs_update = {}
            # store keys to remove after loop in batch-call
            kvs_remove = set()
            for uid, h_int, h_vec in zip(uids, h_ints, h_vectors):
                if h_int not in kvs_update:
                    # First time seeing key, cache current value
                    kvs_update[h_int] = \
                        self.hash2uuids_kvstore.get(h_int, set())
                kvs_update[h_int] -= {uid}
                # If the resolves UID set is empty, flag the key for removal.
                if not kvs_update[h_int]:
                    del kvs_update[h_int]
                    kvs_remove.add(h_int)
                    hashes_for_removal.append(h_vec)
            self._log.debug("Updating hash2uuids: modified relations")
            self.hash2uuids_kvstore.add_many(kvs_update)
            self._log.debug("Updating hash2uuids: removing empty hash keys")
            self.hash2uuids_kvstore.remove_many(kvs_remove)
            del kvs_update, kvs_remove

            # call remove-from-index on hash-index if we have one and there are
            # hashes to be removed.
            if self.hash_index and hashes_for_removal:
                self.hash_index.remove_from_index(hashes_for_removal)

            # Remove descriptors from our set matching the given UIDs.
            self.descriptor_index.remove_many_descriptors(uids)

    def _nn(self, d, n=1):
        """
        Internal method to be implemented by sub-classes to return the nearest
        `N` neighbors to the given descriptor element.

        When this internal method is called, we have already checked that there
        is a vector in ``d`` and our index is not empty.

        :param d: Descriptor element to compute the neighbors of.
        :type d: smqtk.representation.DescriptorElement

        :param n: Number of nearest neighbors to find.
        :type n: int

        :return: Tuple of nearest N DescriptorElement instances, and a tuple of
            the distance values to those neighbors.
        :rtype: (tuple[smqtk.representation.DescriptorElement], tuple[float])

        """
        self._log.debug("generating hash for descriptor")
        d_v = d.vector()
        d_h = self.lsh_functor.get_hash(d_v)

        def comp_descr_dist(d2_v):
            return self._distance_function(d_v, d2_v)

        with self._model_lock:
            self._log.debug("getting near hashes")
            hi = self.hash_index
            if hi is None:
                # Make on-the-fly linear index
                hi = LinearHashIndex()
                # not calling ``build_index`` because we already have the int
                # hashes.
                hi.index = set(self.hash2uuids_kvstore.keys())
            near_hashes, _ = hi.nn(d_h, n)

            self._log.debug("getting UUIDs of descriptors for nearby hashes")
            neighbor_uuids = []
            for h_int in map(bit_vector_to_int_large, near_hashes):
                # If descriptor hash not in our map, we effectively skip it.
                # Get set of descriptor UUIDs for a hash code.
                #: :type: set[collections.Hashable]
                near_uuids = self.hash2uuids_kvstore.get(h_int, set())
                # Accumulate matching descriptor UUIDs to a list.
                neighbor_uuids.extend(near_uuids)
            self._log.debug("-- matched %d UUIDs", len(neighbor_uuids))

            self._log.debug("getting descriptors for neighbor_uuids")
            neighbors = \
                list(self.descriptor_index.get_many_descriptors(neighbor_uuids))

        # Done with model parts at this point, so releasing lock.

        self._log.debug("ordering descriptors via distance method '%s'",
                        self.distance_method)
        self._log.debug('-- getting element vectors')
        neighbor_vectors = elements_to_matrix(neighbors,
                                              report_interval=1.0)
        self._log.debug('-- calculating distances')
        distances = list(map(comp_descr_dist, neighbor_vectors))
        self._log.debug('-- ordering')
        ordered = sorted(zip(neighbors, distances),
                         key=lambda p: p[1])
        self._log.debug('-- slicing top n=%d', n)
        return list(zip(*(ordered[:n])))


# Marking only LSH as the valid impl, otherwise the hash index default would
#   also be picked up (because it also descends from NearestNeighborsIndex).
NN_INDEX_CLASS = LSHNearestNeighborIndex



"""

References:
    [1] http://stackoverflow.com/questions/866465/sql-order-by-the-in-value-list
        answer from "a_horse_with_no_name"

"""
import logging
import multiprocessing

from six.moves import zip
from six.moves import cPickle as pickle

from smqtk.representation import DescriptorIndex
from smqtk.exceptions import ReadOnlyError
from smqtk.utils.postgres \
    import norm_psql_cmd_string, PsqlConnectionHelper

try:
    import psycopg2
except ImportError as ex:
    logging.getLogger(__name__)\
           .warning("Failed to import psycopg2: %s", str(ex))
    psycopg2 = None


PSQL_TABLE_CREATE_RLOCK = multiprocessing.RLock()


# noinspection SqlNoDataSourceInspection
class PostgresDescriptorIndex (DescriptorIndex):
    """
    DescriptorIndex implementation that stored DescriptorElement references in
    a PostgreSQL database.

    A ``PostgresDescriptorIndex`` effectively controls the entire table. Thus
    a ``clear()`` call will remove everything from the table.

    PostgreSQL version support:
        - 9.4

    Table format:
        <uuid col>      TEXT NOT NULL
        <element col>   BYTEA NOT NULL

        <uuid_col> should be the primary key (we assume unique).

    We require that the no column labels not be 'true' for the use of a value
    return shortcut.

    """

    #
    # The following are SQL query templates. The string formatting using {}'s
    # is used to fill in the query before using it in an execute with instance
    # specific values. The ``%()s`` formatting is special for the execute where-
    # by psycopg2 will fill in the values appropriately as specified in a second
    # dictionary argument to ``cursor.execute(query, value_dict)``.
    #
    UPSERT_TABLE_TMPL = norm_psql_cmd_string("""
        CREATE TABLE IF NOT EXISTS {table_name:s} (
          {uuid_col:s} TEXT NOT NULL,
          {element_col:s} BYTEA NOT NULL,
          PRIMARY KEY ({uuid_col:s})
        );
    """)

    SELECT_TMPL = norm_psql_cmd_string("""
        SELECT {col:s}
          FROM {table_name:s}
    """)

    SELECT_LIKE_TMPL = norm_psql_cmd_string("""
        SELECT {element_col:s}
          FROM {table_name:s}
         WHERE {uuid_col:s} like %(uuid_like)s
    """)

    # So we can ensure we get back elements in specified order
    #   - reference [1]
    SELECT_MANY_ORDERED_TMPL = norm_psql_cmd_string("""
        SELECT {table_name:s}.{element_col:s}
          FROM {table_name:s}
          JOIN (
            SELECT *
            FROM unnest(%(uuid_list)s) with ordinality
          ) AS __ordering__ ({uuid_col:s}, {uuid_col:s}_order)
            ON {table_name:s}.{uuid_col:s} = __ordering__.{uuid_col:s}
          ORDER BY __ordering__.{uuid_col:s}_order
    """)

    UPSERT_TMPL = norm_psql_cmd_string("""
        WITH upsert AS (
          UPDATE {table_name:s}
            SET {element_col:s} = %(element_val)s
            WHERE {uuid_col:s} = %(uuid_val)s
            RETURNING *
          )
        INSERT INTO {table_name:s}
          ({uuid_col:s}, {element_col:s})
          SELECT %(uuid_val)s, %(element_val)s
            WHERE NOT EXISTS (SELECT * FROM upsert)
    """)

    DELETE_LIKE_TMPL = norm_psql_cmd_string("""
        DELETE FROM {table_name:s}
              WHERE {uuid_col:s} like %(uuid_like)s
    """)

    DELETE_MANY_TMPL = norm_psql_cmd_string("""
        DELETE FROM {table_name:s}
              WHERE {uuid_col:s} in %(uuid_tuple)s
          RETURNING uid
    """)

    @classmethod
    def is_usable(cls):
        return psycopg2 is not None

    def __init__(self, table_name='descriptor_index', uuid_col='uid',
                 element_col='element',
                 db_name='postgres', db_host=None, db_port=None, db_user=None,
                 db_pass=None, multiquery_batch_size=1000, pickle_protocol=-1,
                 read_only=False, create_table=True):
        """
        Initialize index instance.

        :param table_name: Name of the table to use.
        :type table_name: str

        :param uuid_col: Name of the column containing the UUID signatures.
        :type uuid_col: str

        :param element_col: Name of the table column that will contain
            serialized elements.
        :type element_col: str

        :param db_name: The name of the database to connect to.
        :type db_name: str

        :param db_host: Host address of the Postgres server. If None, we
            assume the server is on the local machine and use the UNIX socket.
            This might be a required field on Windows machines (not tested yet).
        :type db_host: str | None

        :param db_port: Port the Postgres server is exposed on. If None, we
            assume the default port (5423).
        :type db_port: int | None

        :param db_user: Postgres user to connect as. If None, postgres
            defaults to using the current accessing user account name on the
            operating system.
        :type db_user: str | None

        :param db_pass: Password for the user we're connecting as. This may be
            None if no password is to be used.
        :type db_pass: str | None

        :param multiquery_batch_size: For queries that handle sending or
            receiving many queries at a time, batch queries based on this size.
            If this is None, then no batching occurs.

            The advantage of batching is that it reduces the memory impact for
            queries dealing with a very large number of elements (don't have to
            store the full query for all elements in RAM), but the transaction
            will be some amount slower due to splitting the query into multiple
            transactions.
        :type multiquery_batch_size: int | None

        :param pickle_protocol: Pickling protocol to use. We will use -1 by
            default (latest version, probably binary).
        :type pickle_protocol: int

        :param read_only: Only allow read actions against this index.
            Modification actions will throw a ReadOnlyError exceptions.
        :type read_only: bool

        :param create_table: If this instance should try to create the storing
            table before actions are performed against it when not set to be
            read-only. If the configured user does not have sufficient
            permissions to create the table and it does not currently exist, an
            exception will be raised.
        :type create_table: bool

        """
        super(PostgresDescriptorIndex, self).__init__()

        self.table_name = table_name
        self.uuid_col = uuid_col
        self.element_col = element_col

        self.multiquery_batch_size = multiquery_batch_size
        self.pickle_protocol = pickle_protocol
        self.read_only = bool(read_only)
        self.create_table = create_table

        # Checking parameters where necessary
        if self.multiquery_batch_size is not None:
            self.multiquery_batch_size = int(self.multiquery_batch_size)
            assert self.multiquery_batch_size > 0, \
                "A given batch size must be greater than 0 in size " \
                "(given: %d)." % self.multiquery_batch_size
        assert -1 <= self.pickle_protocol <= 2, \
            ("Given pickle protocol is not in the known valid range. Given: %s"
             % self.pickle_protocol)

        self.psql_helper = PsqlConnectionHelper(db_name, db_host, db_port,
                                                db_user, db_pass,
                                                self.multiquery_batch_size,
                                                PSQL_TABLE_CREATE_RLOCK)
        if not self.read_only and self.create_table:
            self.psql_helper.set_table_upsert_sql(
                self.UPSERT_TABLE_TMPL.format(
                    table_name=self.table_name,
                    uuid_col=self.uuid_col,
                    element_col=self.element_col,
                )
            )

    def get_config(self):
        return {
            "table_name": self.table_name,
            "uuid_col": self.uuid_col,
            "element_col": self.element_col,

            "db_name": self.psql_helper.db_name,
            "db_host": self.psql_helper.db_host,
            "db_port": self.psql_helper.db_port,
            "db_user": self.psql_helper.db_user,
            "db_pass": self.psql_helper.db_pass,

            "multiquery_batch_size": self.multiquery_batch_size,
            "pickle_protocol": self.pickle_protocol,
            "read_only": self.read_only,
            "create_table": self.create_table,
        }

    def count(self):
        """
        :return: Number of descriptor elements stored in this index.
        :rtype: int | long
        """
        # Just count UUID column to limit data read.
        q = self.SELECT_TMPL.format(
            col='count(%s)' % self.uuid_col,
            table_name=self.table_name,
        )

        def exec_hook(cur):
            cur.execute(q)

        # There's only going to be one row returned with one element in it.
        return list(self.psql_helper.single_execute(
            exec_hook, yield_result_rows=True
        ))[0][0]

    def clear(self):
        """
        Clear this descriptor index's entries.
        """
        if self.read_only:
            raise ReadOnlyError("Cannot clear a read-only index.")

        q = self.DELETE_LIKE_TMPL.format(
            table_name=self.table_name,
            uuid_col=self.uuid_col,
        )

        def exec_hook(cur):
            cur.execute(q, {'uuid_like': '%'})

        list(self.psql_helper.single_execute(exec_hook))

    def has_descriptor(self, uuid):
        """
        Check if a DescriptorElement with the given UUID exists in this index.

        :param uuid: UUID to query for
        :type uuid: collections.Hashable

        :return: True if a DescriptorElement with the given UUID exists in this
            index, or False if not.
        :rtype: bool

        """
        q = self.SELECT_LIKE_TMPL.format(
            # hacking return value to something simple
            element_col='true',
            table_name=self.table_name,
            uuid_col=self.uuid_col,
        )

        def exec_hook(cur):
            cur.execute(q, {'uuid_like': str(uuid)})

        # Should either yield one or zero rows
        return bool(list(self.psql_helper.single_execute(
            exec_hook, yield_result_rows=True
        )))

    def add_descriptor(self, descriptor):
        """
        Add a descriptor to this index.

        Adding the same descriptor multiple times should not add multiple copies
        of the descriptor in the index (based on UUID). Added descriptors
        overwrite indexed descriptors based on UUID.

        :param descriptor: Descriptor to index.
        :type descriptor: smqtk.representation.DescriptorElement

        """
        if self.read_only:
            raise ReadOnlyError("Cannot clear a read-only index.")

        q = self.UPSERT_TMPL.format(
            table_name=self.table_name,
            uuid_col=self.uuid_col,
            element_col=self.element_col,
        )
        v = {
            'uuid_val': str(descriptor.uuid()),
            'element_val': psycopg2.Binary(
                pickle.dumps(descriptor, self.pickle_protocol)
            )
        }

        def exec_hook(cur):
            cur.execute(q, v)

        list(self.psql_helper.single_execute(exec_hook))

    def add_many_descriptors(self, descriptors):
        """
        Add multiple descriptors at one time.

        Adding the same descriptor multiple times should not add multiple copies
        of the descriptor in the index (based on UUID). Added descriptors
        overwrite indexed descriptors based on UUID.

        :param descriptors: Iterable of descriptor instances to add to this
            index.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """
        if self.read_only:
            raise ReadOnlyError("Cannot clear a read-only index.")

        q = self.UPSERT_TMPL.format(
            table_name=self.table_name,
            uuid_col=self.uuid_col,
            element_col=self.element_col,
        )

        # Transform input into
        def iter_elements():
            for d in descriptors:
                yield {
                    'uuid_val': str(d.uuid()),
                    'element_val': psycopg2.Binary(
                        pickle.dumps(d, self.pickle_protocol)
                    )
                }

        def exec_hook(cur, batch):
            cur.executemany(q, batch)

        self._log.debug("Adding many descriptors")
        list(self.psql_helper.batch_execute(iter_elements(), exec_hook,
                                            self.multiquery_batch_size))

    def get_descriptor(self, uuid):
        """
        Get the descriptor in this index that is associated with the given UUID.

        :param uuid: UUID of the DescriptorElement to get.
        :type uuid: collections.Hashable

        :raises KeyError: The given UUID doesn't associate to a
            DescriptorElement in this index.

        :return: DescriptorElement associated with the queried UUID.
        :rtype: smqtk.representation.DescriptorElement

        """
        q = self.SELECT_LIKE_TMPL.format(
            element_col=self.element_col,
            table_name=self.table_name,
            uuid_col=self.uuid_col,
        )
        v = {'uuid_like': str(uuid)}

        def eh(c):
            c.execute(q, v)
            if c.rowcount == 0:
                raise KeyError(uuid)
            elif c.rowcount != 1:
                raise RuntimeError("Found more than one entry for the given "
                                   "uuid '%s' (got: %d)"
                                   % (uuid, c.rowcount))

        r = list(self.psql_helper.single_execute(eh, yield_result_rows=True))
        return pickle.loads(str(r[0][0]))

    def get_many_descriptors(self, uuids):
        """
        Get an iterator over descriptors associated to given descriptor UUIDs.

        :param uuids: Iterable of descriptor UUIDs to query for.
        :type uuids: collections.Iterable[collections.Hashable]

        :raises KeyError: A given UUID doesn't associate with a
            DescriptorElement in this index.

        :return: Iterator of descriptors associated to given uuid values.
        :rtype: __generator[smqtk.representation.DescriptorElement]

        """
        q = self.SELECT_MANY_ORDERED_TMPL.format(
            table_name=self.table_name,
            element_col=self.element_col,
            uuid_col=self.uuid_col,
        )

        # Cache UUIDs received in order so we can check when we miss one in
        # order to raise a KeyError.
        uuid_order = []

        def iterelems():
            for uid in uuids:
                uuid_order.append(uid)
                yield str(uid)

        def exec_hook(cur, batch):
            v = {'uuid_list': batch}
            # self._log.debug('query: %s', cur.mogrify(q, v))
            cur.execute(q, v)

        self._log.debug("Getting many descriptors")
        # The SELECT_MANY_ORDERED_TMPL query ensures that elements returned are
        #   in the UUID order given to this method. Thus, if the iterated UUIDs
        #   and iterated return rows do not exactly line up, the query join
        #   failed to match a query UUID to something in the database.
        #   - We also check that the number of rows we got back is the same
        #     as elements yielded, else there were trailing UUIDs that did not
        #     match anything in the database.
        g = self.psql_helper.batch_execute(iterelems(), exec_hook,
                                           self.multiquery_batch_size,
                                           yield_result_rows=True)
        i = 0
        for r, expected_uuid in zip(g, uuid_order):
            d = pickle.loads(str(r[0]))
            if d.uuid() != expected_uuid:
                raise KeyError(expected_uuid)
            yield d
            i += 1

        if len(uuid_order) != i:
            # just report the first one that's bad
            raise KeyError(uuid_order[i])

    def remove_descriptor(self, uuid):
        """
        Remove a descriptor from this index by the given UUID.

        :param uuid: UUID of the DescriptorElement to remove.
        :type uuid: collections.Hashable

        :raises KeyError: The given UUID doesn't associate to a
            DescriptorElement in this index.

        """
        if self.read_only:
            raise ReadOnlyError("Cannot remove from a read-only index.")

        q = self.DELETE_LIKE_TMPL.format(
            table_name=self.table_name,
            uuid_col=self.uuid_col,
        )
        v = {'uuid_like': str(uuid)}

        def execute(c):
            c.execute(q, v)
            # Nothing deleted if rowcount == 0
            # (otherwise 1 when deleted a thing)
            if c.rowcount == 0:
                raise KeyError(uuid)

        list(self.psql_helper.single_execute(execute))

    def remove_many_descriptors(self, uuids):
        """
        Remove descriptors associated to given descriptor UUIDs from this index.

        :param uuids: Iterable of descriptor UUIDs to remove.
        :type uuids: collections.Iterable[collections.Hashable]

        :raises KeyError: A given UUID doesn't associate with a
            DescriptorElement in this index.

        """
        if self.read_only:
            raise ReadOnlyError("Cannot remove from a read-only index.")

        q = self.DELETE_MANY_TMPL.format(
            table_name=self.table_name,
            uuid_col=self.uuid_col,
        )
        str_uuid_set = set(str(uid) for uid in uuids)
        v = {'uuid_tuple': tuple(str_uuid_set)}

        def execute(c):
            c.execute(q, v)

            # Check query UUIDs against rows that would actually be deleted.
            deleted_uuid_set = set(r[0] for r in c.fetchall())
            for uid in str_uuid_set:
                if uid not in deleted_uuid_set:
                    raise KeyError(uid)

        list(self.psql_helper.single_execute(execute))

    def iterkeys(self):
        """
        Return an iterator over indexed descriptor keys, which are their UUIDs.
        :rtype: collections.Iterator[collections.Hashable]
        """
        # Getting UUID through the element because the UUID might not be a
        # string type, and the true type is encoded with the DescriptorElement
        # instance.
        for d in self.iterdescriptors():
            yield d.uuid()

    def iterdescriptors(self):
        """
        Return an iterator over indexed descriptor element instances.
        :rtype: collections.Iterator[smqtk.representation.DescriptorElement]
        """
        def execute(c):
            c.execute(self.SELECT_TMPL.format(
                col=self.element_col,
                table_name=self.table_name
            ))

        #: :type: __generator
        execution_results = self.psql_helper.single_execute(
            execute, yield_result_rows=True, named=True
        )
        for r in execution_results:
            d = pickle.loads(str(r[0]))
            yield d

    def iteritems(self):
        """
        Return an iterator over indexed descriptor key and instance pairs.
        :rtype: collections.Iterator[(collections.Hashable,
                                      smqtk.representation.DescriptorElement)]
        """
        for d in self.iterdescriptors():
            yield d.uuid(), d



import itertools
import logging
import multiprocessing
import os
import tempfile

from six.moves import cPickle

import numpy

from smqtk.algorithms.nn_index import NearestNeighborsIndex
from smqtk.representation.data_element import from_uri
from smqtk.representation.descriptor_element import elements_to_matrix

# Requires FLANN bindings
try:
    import pyflann
except ImportError:
    pyflann = None


class FlannNearestNeighborsIndex (NearestNeighborsIndex):
    """
    Nearest-neighbor computation using the FLANN library (pyflann module).

    This implementation uses in-memory data structures, and thus has an index
    size limit based on how much memory the running machine has available.

    NOTE ON MULTIPROCESSING
        Normally, FLANN indices don't play well when multiprocessing due to
        the underlying index being a C structure, which doesn't auto-magically
        transfer to forked processes like python structure data does. The
        serialized FLANN index file is used to restore a built index in
        separate processes, assuming one has been built.

    """

    @classmethod
    def is_usable(cls):
        # if underlying library is not found, the import above will error
        return pyflann is not None

    def __init__(self, index_uri=None, parameters_uri=None,
                 descriptor_cache_uri=None,
                 # Parameters for building an index
                 autotune=False, target_precision=0.95, sample_fraction=0.1,
                 distance_method='hik', random_seed=None):
        """
        Initialize FLANN index properties. Does not contain a query-able index
        until one is built via the ``build_index`` method, or loaded from
        existing model files.

        When using this algorithm in a multiprocessing environment, the model
        file path parameters must be specified due to needing to reload the
        FLANN index on separate processes. This is because FLANN is in C and
        its instances are not copied into processes.

        Documentation on index building parameters and their meaning can be
        found in the FLANN documentation PDF:

            http://www.cs.ubc.ca/research/flann/uploads/FLANN/flann_manual-1.8.4.pdf

        See the MATLAB section for detailed descriptions (python section will
        just point you to the MATLAB section).

        :param index_uri: Optional URI to where to load/store FLANN index
            when initialized and/or built.

            If not configured, no model files are written to or loaded from
            disk.
        :type index_uri: None | str

        :param parameters_uri: Optional file location to load/save FLANN
            index parameters determined at build time.

            If not configured, no model files are written to or loaded from
            disk.
        :type parameters_uri: None | str

        :param descriptor_cache_uri: Optional file location to load/store
            DescriptorElements in this index.

            If not configured, no model files are written to or loaded from
            disk.
        :type descriptor_cache_uri: None | str

        :param autotune: Whether or not to perform parameter auto-tuning when
            building the index. If this is False, then the `target_precision`
            and `sample_fraction` parameters are not used.
        :type autotune: bool

        :param target_precision: Target estimation accuracy when determining
            nearest neighbor when tuning parameters. This should be between
            [0,1] and represents percentage accuracy.
        :type target_precision: float

        :param sample_fraction: Sub-sample percentage of the total index to use
            when performing auto-tuning. Value should be in the range of [0,1]
            and represents percentage.
        :type sample_fraction: float

        :param distance_method: Method label of the distance function to use.
            See FLANN documentation manual for available methods. Common
            methods include "hik", "chi_square" (default), and "euclidean".
            When loading and existing index, this value is ignored in
            preference for the distance method used to build the loaded index.
        :type distance_method: str

        :param random_seed: Integer to use as the random number generator seed.
        :type random_seed: int

        """
        super(FlannNearestNeighborsIndex, self).__init__()

        self._index_uri = index_uri
        self._index_param_uri = parameters_uri
        self._descr_cache_uri = descriptor_cache_uri

        # Elements will be None if input URI is None
        self._index_elem = \
            self._index_uri and from_uri(self._index_uri)
        self._index_param_elem = \
            self._index_param_uri and from_uri(self._index_param_uri)
        self._descr_cache_elem = \
            self._descr_cache_uri and from_uri(self._descr_cache_uri)

        # parameters for building an index
        self._build_autotune = autotune
        self._build_target_precision = float(target_precision)
        self._build_sample_frac = float(sample_fraction)
        self._distance_method = str(distance_method)

        # Lock for model component access.  Using a multiprocessing due to
        # possible cases where another thread/process attempts to restore a
        # model before its fully written.  A reordering of _build_index could
        # lessen the requirement to a `threading.RLock`.
        self._model_lock = multiprocessing.RLock()

        # In-order cache of descriptors we're indexing over.
        # - flann.nn_index will spit out indices to list
        #: :type: list[smqtk.representation.DescriptorElement] | None
        self._descr_cache = None

        # The flann instance with a built index. None before index load/build.
        #: :type: pyflann.index.FLANN or None
        self._flann = None
        # Flann index parameters determined during building. None before index
        # load/build.
        #: :type: dict
        self._flann_build_params = None

        #: :type: None | int
        self._rand_seed = None
        if random_seed:
            self._rand_seed = int(random_seed)

        # The process ID that the currently set FLANN instance was built/loaded
        # on. If this differs from the current process ID, the index should be
        # reloaded from cache.
        self._pid = None

        # Load the index/parameters if one exists
        if self._has_model_data():
            self._log.info("Found existing model data. Loading.")
            self._load_flann_model()

    def get_config(self):
        return {
            "index_uri": self._index_uri,
            "parameters_uri": self._index_param_uri,
            "descriptor_cache_uri": self._descr_cache_uri,
            "autotune": self._build_autotune,
            "target_precision": self._build_target_precision,
            "sample_fraction": self._build_sample_frac,
            "distance_method": self._distance_method,
            "random_seed": self._rand_seed,
        }

    def _has_model_data(self):
        """
        check if configured model files are configured and not empty
        """
        return (self._index_elem and not self._index_elem.is_empty() and
                self._index_param_elem and
                not self._index_param_elem.is_empty() and
                self._descr_cache_elem and
                not self._descr_cache_elem.is_empty())

    def _load_flann_model(self):
        if not self._descr_cache and not self._descr_cache_elem.is_empty():
            # Load descriptor cache
            # - is copied on fork, so only need to load here.
            self._log.debug("Loading cached descriptors")
            self._descr_cache = \
                cPickle.loads(self._descr_cache_elem.get_bytes())

        # Params pickle include the build params + our local state params
        if self._index_param_elem and not self._index_param_elem.is_empty():
            state = cPickle.loads(self._index_param_elem.get_bytes())
            self._build_autotune = state['b_autotune']
            self._build_target_precision = state['b_target_precision']
            self._build_sample_frac = state['b_sample_frac']
            self._distance_method = state['distance_method']
            self._flann_build_params = state['flann_build_params']

        # Load the binary index
        if self._index_elem and not self._index_elem.is_empty():
            # make numpy matrix of descriptor vectors for FLANN
            pts_array = [d.vector() for d in self._descr_cache]
            pts_array = numpy.array(pts_array, dtype=pts_array[0].dtype)
            pyflann.set_distance_type(self._distance_method)
            self._flann = pyflann.FLANN()
            tmp_fp = self._index_elem.write_temp()
            self._flann.load_index(tmp_fp, pts_array)
            self._index_elem.clean_temp()
            del pts_array, tmp_fp

        # Set current PID to the current
        self._pid = multiprocessing.current_process().pid

    def _restore_index(self):
        """
        If we think we're suppose to have an index, check the recorded PID with
        the current PID, reloading the index from cache if they differ.

        If there is a loaded index and we're on the same process that created
        it this does nothing.
        """
        if bool(self._flann) \
                and self._has_model_data() \
                and self._pid != multiprocessing.current_process().pid:
            self._load_flann_model()

    def count(self):
        """
        :return: Number of elements in this index.
        :rtype: int
        """
        with self._model_lock:
            if self._descr_cache is not None:
                return len(self._descr_cache)
            else:
                return 0

    def _build_index(self, descriptors):
        """
        Internal method to be implemented by sub-classes to build the index
        with the given descriptor data elements.

        Subsequent calls to this method should rebuild the current index.  This
        method shall not add to the existing index nor raise an exception to as
        to protect the current index.

        Implementation Notes:
            - We keep a cache file serialization around for our index in case
                sub-processing occurs so as to be able to recover from the
                underlying C data not being there. This could cause issues if
                a main or child process rebuild's the index, as we clear the
                old cache away.

        :param descriptors: Iterable of descriptor elements to build index
            over.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """
        with self._model_lock:
            # Not caring about restoring the index because we're just making a
            # new one.
            self._log.info("Building new FLANN index")

            self._log.debug("Caching descriptor elements")
            self._descr_cache = list(descriptors)
            # Cache descriptors if we have an element
            if self._descr_cache_elem and self._descr_cache_elem.writable():
                self._log.debug("Caching descriptors: %s",
                                self._descr_cache_elem)
                self._descr_cache_elem.set_bytes(
                    cPickle.dumps(self._descr_cache, -1)
                )

            params = {
                "target_precision": self._build_target_precision,
                "sample_fraction": self._build_sample_frac,
                "log_level": ("info"
                              if self._log.getEffectiveLevel() <= logging.DEBUG
                              else "warning")
            }
            if self._build_autotune:
                params['algorithm'] = "autotuned"
            if self._rand_seed is not None:
                params['random_seed'] = self._rand_seed
            pyflann.set_distance_type(self._distance_method)

            self._log.debug("Accumulating descriptor vectors into matrix for "
                            "FLANN")
            pts_array = elements_to_matrix(self._descr_cache,
                                           report_interval=1.0)

            self._log.debug('Building FLANN index')
            self._flann = pyflann.FLANN()
            self._flann_build_params = self._flann.build_index(pts_array,
                                                               **params)
            del pts_array

            if self._index_elem and self._index_elem.writable():
                self._log.debug("Caching index: %s", self._index_elem)
                # FLANN wants to write to a file, so make a temp file, then
                # read it in, putting bytes into element.
                fd, fp = tempfile.mkstemp()
                try:
                    self._flann.save_index(fp)
                    # Use the file descriptor to create the file object.
                    # This avoids reopening the file and will automatically
                    # close the file descriptor on exiting the with block.
                    # fdopen() is required because in Python 2 open() does
                    # not accept a file descriptor.
                    with os.fdopen(fd, 'rb') as f:
                        self._index_elem.set_bytes(f.read())
                finally:
                    os.remove(fp)
            if self._index_param_elem and self._index_param_elem.writable():
                self._log.debug("Caching index params: %s",
                                self._index_param_elem)
                state = {
                    'b_autotune': self._build_autotune,
                    'b_target_precision': self._build_target_precision,
                    'b_sample_frac': self._build_sample_frac,
                    'distance_method': self._distance_method,
                    'flann_build_params': self._flann_build_params,
                }
                self._index_param_elem.set_bytes(cPickle.dumps(state, -1))

            self._pid = multiprocessing.current_process().pid

    def _update_index(self, descriptors):
        """
        Internal method to be implemented by sub-classes to additively update
        the current index with the one or more descriptor elements given.

        If no index exists yet, a new one should be created using the given
        descriptors.

        The currently bundled FLANN implementation bindings (v1.8.4) does not
        support support incremental updating of an existing index.  Thus this
        update method fully rebuilds the index based on the previous cache of
        descriptors and the newly specified ones.  Due to requiring a full
        rebuild this update method may take a significant amount of time
        depending on the size of the index being updated.

        :param descriptors: Iterable of descriptor elements to add to this
            index.
        :type descriptors: collections.Iterable[smqtk.representation
                                                     .DescriptorElement]

        """
        with self._model_lock:
            self._restore_index()
            # Build a new index that contains the union of the current
            # descriptors and the new provided descriptors.
            self._log.info("Rebuilding FLANN index to include new "
                           "descriptors.")
            self.build_index(itertools.chain(self._descr_cache, descriptors))

    def _remove_from_index(self, uids):
        """
        Internal method to be implemented by sub-classes to partially remove
        descriptors from this index associated with the given UIDs.

        :param uids: Iterable of UIDs of descriptors to remove from this index.
        :type uids: collections.Iterable[collections.Hashable]

        :raises KeyError: One or more UIDs provided do not match any stored
            descriptors.

        """
        with self._model_lock:
            self._restore_index()
            uidset = set(uids)
            # Make sure provided UIDs are part of our current index.
            uid_diff = uidset - set(map(lambda d: d.uuid(), self._descr_cache))
            if uid_diff:
                if len(uid_diff) == 1:
                    raise KeyError(list(uid_diff)[0])
                else:
                    raise KeyError(uid_diff)
            # Filter descriptors NOT matching UIDs of current descriptor
            # cache.
            self._descr_cache = \
                [d for d in self._descr_cache if d.uuid() in uidset]
            self.build_index(self._descr_cache)

    def _nn(self, d, n=1):
        """
        Internal method to be implemented by sub-classes to return the nearest
        `N` neighbors to the given descriptor element.

        When this internal method is called, we have already checked that there
        is a vector in ``d`` and our index is not empty.

        :param d: Descriptor element to compute the neighbors of.
        :type d: smqtk.representation.DescriptorElement

        :param n: Number of nearest neighbors to find.
        :type n: int

        :return: Tuple of nearest N DescriptorElement instances, and a tuple of
            the distance values to those neighbors.
        :rtype: (tuple[smqtk.representation.DescriptorElement], tuple[float])

        """
        with self._model_lock:
            self._restore_index()
            vec = d.vector()

            # If the distance method is HIK, we need to treat it special since
            # that method produces a similarity score, not a distance score.
            #
            # FLANN asserts that we query for <= index size, thus the use of
            # min().
            if self._distance_method == 'hik':
                # This call is different than the else version in that k is the
                # size of the full data set, so that we can reverse the
                # distances.
                #: :type: numpy.ndarray, numpy.ndarray
                idxs, dists = self._flann.nn_index(
                    vec, len(self._descr_cache),
                    **self._flann_build_params
                )
            else:
                #: :type: numpy.ndarray, numpy.ndarray
                idxs, dists = self._flann.nn_index(
                    vec, min(n, len(self._descr_cache)),
                    **self._flann_build_params
                )

            # When N>1, return value is a 2D array. Since this method limits
            # query to a single descriptor, we reduce to 1D arrays.
            if len(idxs.shape) > 1:
                idxs = idxs[0]
                dists = dists[0]

            if self._distance_method == 'hik':
                # Invert values to stay consistent with other distance value
                # norms.
                dists = [1.0 - d for d in dists]
                idxs = tuple(reversed(idxs))[:n]
                dists = tuple(reversed(dists))[:n]
            else:
                dists = tuple(dists)

            return [self._descr_cache[i] for i in idxs], dists


NN_INDEX_CLASS = FlannNearestNeighborsIndex



from __future__ import division, print_function
import mock
import os
import six
import unittest

from smqtk.representation.data_element.memory_element import DataMemoryElement
from smqtk.representation.data_set.file_set import DataFileSet
from smqtk.tests import TEST_DATA_DIR


class TestDataFileSet (unittest.TestCase):

    def test_configuration(self):
        default_config = DataFileSet.get_default_config()
        self.assertEqual(default_config, {
            'root_directory': None,
            'uuid_chunk': 10,
            'pickle_protocol': -1,
        })

        default_config['root_directory'] = '/some/dir'
        inst1 = DataFileSet.from_config(default_config)
        self.assertEqual(default_config, inst1.get_config())

        inst2 = DataFileSet.from_config(inst1.get_config())
        self.assertEqual(inst1, inst2)

    def test_new_invalid_uuid_chunk(self):
        self.assertRaises(
            ValueError,
            DataFileSet, '/', 0
        )

        self.assertRaises(
            ValueError,
            DataFileSet, '/', -1
        )

    def test_new(self):
        # The following should be valid constructor parameter setups
        DataFileSet('/')
        DataFileSet('/', uuid_chunk=None)
        DataFileSet('/', uuid_chunk=1)
        DataFileSet('/', uuid_chunk=2346)

        DataFileSet('relative/path')
        DataFileSet('relative/path', uuid_chunk=None)
        DataFileSet('relative/path', uuid_chunk=1)
        DataFileSet('relative/path', uuid_chunk=2346)

    def test_iter_file_tree_chunk0(self):
        test_dir_path = os.path.join(TEST_DATA_DIR, 'test_data_file_set_tree')
        expected_filepaths = {
            os.path.join(test_dir_path, 'UUID_0.dataElement'),
            os.path.join(test_dir_path, 'UUID_1.dataElement'),
            os.path.join(test_dir_path, 'UUID_2.dataElement'),
        }

        dfs = DataFileSet(test_dir_path, uuid_chunk=None, pickle_protocol=2)
        actual_filepaths = set(dfs._iter_file_tree())
        self.assertSetEqual(actual_filepaths, expected_filepaths)

    def test_iter_file_tree_chunk3(self):
        test_dir_path = os.path.join(TEST_DATA_DIR, 'test_data_file_set_tree')
        expected_filepaths = {
            os.path.join(test_dir_path, '0/0/UUID_000.dataElement'),
            os.path.join(test_dir_path, '0/0/UUID_001.dataElement'),
            os.path.join(test_dir_path, '0/1/UUID_012.dataElement'),
            os.path.join(test_dir_path, '1/8/UUID_180.dataElement'),
            os.path.join(test_dir_path, '3/1/UUID_317.dataElement'),
        }

        dfs = DataFileSet(test_dir_path, uuid_chunk=3, pickle_protocol=2)
        actual_filepaths = set(dfs._iter_file_tree())
        self.assertSetEqual(actual_filepaths, expected_filepaths)

    def test_iter_file_tree_chunk4(self):
        test_dir_path = os.path.join(TEST_DATA_DIR, 'test_data_file_set_tree')
        expected_filepaths = {
            os.path.join(test_dir_path, '4/3/2/1/UUID_43210.dataElement'),
        }

        dfs = DataFileSet(test_dir_path, uuid_chunk=5, pickle_protocol=2)
        actual_filepaths = set(dfs._iter_file_tree())
        self.assertSetEqual(actual_filepaths, expected_filepaths)

    def test_containing_dir_str_uuid(self):
        # Chunk == None
        s = DataFileSet('/', uuid_chunk=None)
        self.assertEqual(s._containing_dir('0000'), '/')
        self.assertEqual(s._containing_dir('346'), '/')
        # Chunk == 1
        s = DataFileSet('/', uuid_chunk=1)
        self.assertEqual(s._containing_dir('0000'), '/')
        self.assertEqual(s._containing_dir('346'), '/')
        # Chunk == 3
        s = DataFileSet('/', uuid_chunk=3)
        self.assertEqual(s._containing_dir('123456'), '/12/34')
        self.assertEqual(s._containing_dir('685225624578'), '/6852/2562')
        self.assertEqual(s._containing_dir('1234567'), '/123/45')

    def test_containing_dir_not_str_uuid(self):
        self.assertEqual(
            DataFileSet('/', None)._containing_dir(4123458),
            "/"
        )
        self.assertEqual(
            DataFileSet('/', 3)._containing_dir(4123458),
            "/412/34"
        )

    def test_fp_for_uuid(self):
        self.assertEqual(
            DataFileSet('/', None)._fp_for_uuid(0),
            '/UUID_0.dataElement'
        )
        self.assertEqual(
            DataFileSet('/', None)._fp_for_uuid('abc'),
            '/UUID_abc.dataElement'
        )
        self.assertEqual(
            DataFileSet('/', 3)._fp_for_uuid('abc'),
            '/a/b/UUID_abc.dataElement'
        )

    @mock.patch('smqtk.representation.data_set.file_set.pickle')
    @mock.patch('smqtk.representation.data_set.file_set.open',
                new_callable=mock.MagicMock)
    def test_iter(self, m_open, m_pickle):
        expected_file_tree_iter = ['/a', '/b', '/d']
        dfs = DataFileSet('/')
        dfs._iter_file_tree = mock.MagicMock(
            return_value=expected_file_tree_iter)
        list(dfs)
        self.assertEqual(m_open.call_count, 3)
        self.assertEqual(m_pickle.load.call_count, 3)
        m_open.assert_any_call('/a')
        m_open.assert_any_call('/b')
        m_open.assert_any_call('/d')

    def test_count(self):
        expected_file_tree_iter = ['/a', '/b', '/d']
        dfs = DataFileSet('/')
        dfs._iter_file_tree = mock.MagicMock(
            return_value=expected_file_tree_iter)
        self.assertEqual(dfs.count(), 3)

    def test_uuids(self):
        # mocking self iteration results
        expected_data_elements = [
            DataMemoryElement(six.b('a')),
            DataMemoryElement(six.b('b')),
            DataMemoryElement(six.b('v')),
        ]
        expected_uuid_set = {
            DataMemoryElement(six.b('a')).uuid(),
            DataMemoryElement(six.b('b')).uuid(),
            DataMemoryElement(six.b('v')).uuid(),
        }

        # Replacement iterator for DataFileSet to yield expected test values.
        def test_iter():
            for e in expected_data_elements:
                yield e

        with mock.patch('smqtk.representation.data_set.file_set.DataFileSet'
                        '.__iter__') as m_iter:
            m_iter.side_effect = test_iter

            dfs = DataFileSet('/')
            self.assertSetEqual(dfs.uuids(), expected_uuid_set)

    def test_add_data_not_dataelement(self):
        dfs = DataFileSet('/')
        self.assertRaisesRegexp(
            AssertionError,
            "^Not given a DataElement for addition:",
            dfs.add_data, 'not a dataElement'
        )

    @mock.patch('smqtk.representation.data_set.file_set.pickle')
    @mock.patch('smqtk.representation.data_set.file_set.open')
    @mock.patch('smqtk.representation.data_set.file_set.file_utils'
                '.safe_create_dir')
    @mock.patch('smqtk.representation.data_set.file_set.isinstance')
    def test_add_data_single(self, m_isinstance, m_scd, m_open, _m_pickle):
        # Pretend that we are giving DataElement instances
        m_isinstance.return_value = True

        # Testing that appropriate directories are given to safe_create_dir and
        # appropriate filepaths are passed to open.
        expected_uuid = 'abcd'

        mock_elem = mock.MagicMock()
        mock_elem.uuid.return_value = expected_uuid

        dfs = DataFileSet('/', uuid_chunk=None)
        dfs.add_data(mock_elem)
        m_scd.assert_called_with('/')
        m_open.assert_called_with('/UUID_abcd.dataElement', 'wb')

        dfs = DataFileSet('/', uuid_chunk=1)
        dfs.add_data(mock_elem)
        m_scd.assert_called_with('/')
        m_open.assert_called_with('/UUID_abcd.dataElement', 'wb')

        dfs = DataFileSet('/', uuid_chunk=2)
        dfs.add_data(mock_elem)
        m_scd.assert_called_with('/ab')
        m_open.assert_called_with('/ab/UUID_abcd.dataElement', 'wb')

    @mock.patch('smqtk.representation.data_set.file_set.pickle')
    @mock.patch('smqtk.representation.data_set.file_set.open')
    @mock.patch('smqtk.representation.data_set.file_set.file_utils'
                '.safe_create_dir')
    @mock.patch('smqtk.representation.data_set.file_set.isinstance')
    def test_add_data_multiple_chunk0(self, m_isinstance, m_scd, m_open,
                                      _m_pickle):
        # Pretend that we are giving DataElement instances
        m_isinstance.return_value = True

        # Testing that appropriate directories are given to safe_create_dir and
        # appropriate filepaths are passed to open.
        expected_uuid_1 = "abcdefg"
        expected_uuid_2 = "1234567"
        expected_uuid_3 = "4F*s93#5"

        mock_elem_1 = mock.MagicMock()
        mock_elem_1.uuid.return_value = expected_uuid_1
        mock_elem_2 = mock.MagicMock()
        mock_elem_2.uuid.return_value = expected_uuid_2
        mock_elem_3 = mock.MagicMock()
        mock_elem_3.uuid.return_value = expected_uuid_3

        # Chunk = None
        dfs = DataFileSet('/', uuid_chunk=None)
        dfs.add_data(mock_elem_1, mock_elem_2, mock_elem_3)
        # Created root 3 times
        self.assertEqual(m_scd.call_count, 3)
        m_scd.assert_called_with('/')
        # called open correctly 3 times
        self.assertEqual(m_open.call_count, 3)
        m_open.assert_any_call('/UUID_abcdefg.dataElement', 'wb')
        m_open.assert_any_call('/UUID_1234567.dataElement', 'wb')
        m_open.assert_any_call('/UUID_4F*s93#5.dataElement', 'wb')

    @mock.patch('smqtk.representation.data_set.file_set.pickle')
    @mock.patch('smqtk.representation.data_set.file_set.open')
    @mock.patch('smqtk.representation.data_set.file_set.file_utils'
                '.safe_create_dir')
    @mock.patch('smqtk.representation.data_set.file_set.isinstance')
    def test_add_data_multiple_chunk3(self, m_isinstance, m_scd, m_open,
                                      _m_pickle):
        # Pretend that we are giving DataElement instances
        m_isinstance.return_value = True

        # Testing that appropriate directories are given to safe_create_dir and
        # appropriate filepaths are passed to open.
        expected_uuid_1 = "abcdefg"
        expected_uuid_2 = "1234567"
        expected_uuid_3 = "4F*s93#5"

        mock_elem_1 = mock.MagicMock()
        mock_elem_1.uuid.return_value = expected_uuid_1
        mock_elem_2 = mock.MagicMock()
        mock_elem_2.uuid.return_value = expected_uuid_2
        mock_elem_3 = mock.MagicMock()
        mock_elem_3.uuid.return_value = expected_uuid_3

        # Chunk = 3
        dfs = DataFileSet('/', uuid_chunk=3)
        dfs.add_data(mock_elem_1, mock_elem_2, mock_elem_3)
        # Created correct directories
        self.assertEqual(m_scd.call_count, 3)
        m_scd.assert_any_call('/abc/de')
        m_scd.assert_any_call('/123/45')
        m_scd.assert_any_call('/4F*/s93')
        # called open correctly 3 times
        self.assertEqual(m_open.call_count, 3)
        m_open.assert_any_call('/abc/de/UUID_abcdefg.dataElement', 'wb')
        m_open.assert_any_call('/123/45/UUID_1234567.dataElement', 'wb')
        m_open.assert_any_call('/4F*/s93/UUID_4F*s93#5.dataElement', 'wb')

    @mock.patch('smqtk.representation.data_set.file_set.pickle')
    @mock.patch('smqtk.representation.data_set.file_set.open')
    @mock.patch('smqtk.representation.data_set.file_set.file_utils'
                '.safe_create_dir')
    @mock.patch('smqtk.representation.data_set.file_set.isinstance')
    def test_add_data_multiple_chunk3_relative(self, m_isinstance, m_scd,
                                               m_open, _m_pickle):
        # Pretend that we are giving DataElement instances
        m_isinstance.return_value = True

        # Testing that appropriate directories are given to safe_create_dir and
        # appropriate filepaths are passed to open.
        expected_uuid_1 = "abcdefg"
        expected_uuid_2 = "1234567"
        expected_uuid_3 = "4F*s93#5"

        mock_elem_1 = mock.MagicMock()
        mock_elem_1.uuid.return_value = expected_uuid_1
        mock_elem_2 = mock.MagicMock()
        mock_elem_2.uuid.return_value = expected_uuid_2
        mock_elem_3 = mock.MagicMock()
        mock_elem_3.uuid.return_value = expected_uuid_3

        # Chunk = 3
        dfs = DataFileSet('rel/subdir', uuid_chunk=3)
        dfs.add_data(mock_elem_1, mock_elem_2, mock_elem_3)
        # Created correct directories
        self.assertEqual(m_scd.call_count, 3)
        m_scd.assert_any_call('rel/subdir/abc/de')
        m_scd.assert_any_call('rel/subdir/123/45')
        m_scd.assert_any_call('rel/subdir/4F*/s93')
        # called open correctly 3 times
        self.assertEqual(m_open.call_count, 3)
        m_open.assert_any_call('rel/subdir/abc/de/UUID_abcdefg.dataElement',
                               'wb')
        m_open.assert_any_call('rel/subdir/123/45/UUID_1234567.dataElement',
                               'wb')
        m_open.assert_any_call('rel/subdir/4F*/s93/UUID_4F*s93#5.dataElement',
                               'wb')

    @mock.patch('smqtk.representation.data_set.file_set.osp.isfile')
    def test_get_data_no_file(self, m_isfile):
        # Testing when we generate a filepath that does not point to an existing
        # file, meaning the UUID is referring to a dataElement not a part of our
        # set.
        m_isfile.return_value = False

        dfs = DataFileSet(TEST_DATA_DIR, None)
        self.assertRaisesRegexp(
            KeyError,
            'no_exist_uuid',
            dfs.get_data, 'no_exist_uuid'
        )

    @mock.patch('smqtk.representation.data_set.file_set.pickle')
    @mock.patch('smqtk.representation.data_set.file_set.open')
    @mock.patch('smqtk.representation.data_set.file_set.osp.isfile')
    def test_get_data_valid_filepath(self, m_isfile, m_open, m_pickle):
        # Testing that filepath we get back from _fp_for_uuid generator is
        # valid, meaning that the given UUID does refer to a serialized
        # DataElement in our set, which is then opened and returned.
        m_isfile.return_value = True

        expected_uuid = 'abc'
        expected_filepath = os.path.join(TEST_DATA_DIR, 'UUID_abc.dataElement')
        expected_pickle_return = 'loaded DataElement instance'

        m_pickle.load.return_value = expected_pickle_return

        dfs = DataFileSet(TEST_DATA_DIR, None)
        actual_return = dfs.get_data(expected_uuid)
        m_isfile.assert_called_once_with(expected_filepath)
        m_open.assert_called_once_with(expected_filepath, 'rb')
        self.assertEqual(actual_return, expected_pickle_return)



from smqtk.tests.utils.test_plugin_get import DummyInterface


__author__ = 'paul.tunison@kitware.com'


class ImplExternal3 (DummyInterface):

    @classmethod
    def is_usable(cls):
        return True

    def inst_method(self, val):
        return 'external3'+str(val)



"""
LICENCE
-------
Copyright 2013-2016 by Kitware, Inc. All Rights Reserved. Please refer to
KITWARE_LICENSE.TXT for licensing information, or contact General Counsel,
Kitware, Inc., 28 Corporate Drive, Clifton Park, NY 12065.

"""

import copy
import functools
import logging
import operator as op

# noinspection PyUnresolvedReferences
from six.moves import range


class SmqtkObject (object):
    """
    Highest level object interface for classes defined in SMQTK.

    Currently defines logging methods.

    """

    @classmethod
    def get_logger(cls):
        """
        :return: logging object for this class
        :rtype: logging.Logger
        """
        return logging.getLogger('.'.join((cls.__module__, cls.__name__)))

    @property
    def _log(self):
        """
        :return: logging object for this class as a property
        :rtype: logging.Logger
        """
        return self.get_logger()


def ncr(n, r):
    """
    N-choose-r method, returning the number of combinations possible in integer
    form.

    From dheerosaur:
        http://stackoverflow.com/questions/4941753/is-there-a-math-ncr-function-in-python

    :param n: Selection pool size.
    :type n: int

    :param r: permutation selection size.
    :type r: int

    :return: Number of n-choose-r permutations for the given n and r.
    :rtype: int

    """
    r = min(r, n - r)
    if r == 0:
        return 1
    numer = functools.reduce(op.mul, range(n, n - r, -1))
    denom = functools.reduce(op.mul, range(1, r + 1))
    return numer // denom


def merge_dict(a, b, deep_copy=False):
    """
    Merge dictionary b into dictionary a.

    This is different than normal dictionary update in that we don't bash
    nested dictionaries, instead recursively updating them.

    For congruent keys, values are are overwritten, while new keys in ``b`` are
    simply added to ``a``.

    Values are assigned (not copied) by default. Setting ``deep_copy`` causes
    values from ``b`` to be deep-copied into ``a``.

    :param a: The "base" dictionary that is updated in place.
    :type a: dict

    :param b: The dictionary to merge into ``a`` recursively.
    :type b: dict

    :param deep_copy: Optionally deep-copy values from ``b`` when assigning into
        ``a``.
    :type deep_copy: bool

    :return: ``a`` dictionary after merger (not a copy).
    :rtype: dict

    """
    for k in b:
        if k in a and isinstance(a[k], dict) and isinstance(b[k], dict):
            merge_dict(a[k], b[k], deep_copy)
        elif deep_copy:
            a[k] = copy.deepcopy(b[k])
        else:
            a[k] = b[k]
    return a


###
# In specific ordering for dependency resolution
#

# No internal util dependencies
from .bin_utils import initialize_logging
from .configurable_interface import Configurable
from .database_info import DatabaseInfo
from .iter_validation import check_empty_iterable
from .read_write_lock import ReaderUpdateException, DummyRWLock, ReadWriteLock
from .safe_config_comment_parser import SafeConfigCommentParser
from .signal_handler import SignalHandler
from .simple_timer import SimpleTimer



from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

import unittest

import numpy as np

from smqtk.utils import prob_utils


class TestAdjustProba (unittest.TestCase):

    def test_single_class(self):

        num = 10
        dim = 1

        proba = np.random.rand(num, dim)
        self.assertTrue(np.allclose(1, prob_utils.adjust_proba(proba, [1])))

        proba = np.random.rand(num, dim)
        self.assertTrue(np.allclose(1, prob_utils.adjust_proba(proba, [-1])))

        proba = np.ones_like(proba)
        self.assertTrue(np.allclose(1, prob_utils.adjust_proba(proba, [1])))

    def test_failure_cases(self):

        num = 10
        dim = 3

        proba = np.zeros((num, dim))
        self.assertRaisesRegexp(ValueError,
                                "At least one probability must be positive",
                                prob_utils.adjust_proba, proba, [1, 2, 3])

        proba[1] = -1.0
        proba[2] = 1.0
        self.assertRaisesRegexp(ValueError,
                                "Probabilities must be at least 0",
                                prob_utils.adjust_proba, proba, [1, 2, 3])

    def test_shape_cases(self):

        num = 10
        dim = 3

        proba = np.random.rand(num, dim)
        self.assertRaisesRegexp(ValueError,
                                "The dimensions of probabilities and "
                                "adjustments must be compatible.",
                                prob_utils.adjust_proba, proba, [1, 2])

        proba = np.random.rand(1, dim)
        proba /= proba.sum()
        self.assertTrue(
            np.allclose(proba, prob_utils.adjust_proba(proba, [1, 1, 1])))

        self.assertRaisesRegexp(ValueError,
                                "The dimensions of probabilities and "
                                "adjustments must be compatible.",
                                prob_utils.adjust_proba,
                                np.ones((num, 1)), np.ones((1, num)))

    def test_adjust_constant(self):

        num = 10
        dim = 3

        proba = np.random.rand(num, dim)
        proba /= proba.sum(axis=1, keepdims=True)

        self.assertTrue(
            np.allclose(proba, prob_utils.adjust_proba(proba, [1, 1, 1])))
        self.assertTrue(
            np.allclose(proba, prob_utils.adjust_proba(proba, [10, 10, 10])))

    def test_adjust_serial_vs_sum(self):

        num = 10
        dim = 3

        proba = np.random.rand(num, dim)
        proba /= proba.sum(axis=1, keepdims=True)

        adj1 = np.array([1, 2, 3])
        adj2 = np.array([2, 0, -2])

        proba_fst = prob_utils.adjust_proba(proba, adj1)
        proba_snd = prob_utils.adjust_proba(proba_fst, adj2)
        proba_sum = prob_utils.adjust_proba(proba, adj1 + adj2)

        self.assertTrue(np.allclose(proba_snd, proba_sum))

        proba_fst = prob_utils.adjust_proba(proba, adj1)
        proba_snd = prob_utils.adjust_proba(proba_fst, -adj1)

        self.assertTrue(np.allclose(proba_snd, proba))

    def test_adjust(self):

        num = 10
        dim = 3

        proba = np.random.rand(num, dim)
        proba /= proba.sum(axis=1, keepdims=True)

        adj = [0, 1, 0]

        proba_post = prob_utils.adjust_proba(proba, adj)
        comp = proba_post > proba
        self.assertTrue(np.all([False, True, False] == comp))
        comp = proba_post < proba
        self.assertTrue(np.all([True, False, True] == comp))
        comp = np.isclose(proba, proba_post)
        self.assertFalse(np.any(comp))

        adj = [-1, 0, 0]

        proba_post = prob_utils.adjust_proba(proba, adj)
        comp = proba_post < proba
        self.assertTrue(np.all([True, False, False] == comp))
        comp = proba_post > proba
        self.assertTrue(np.all([False, True, True] == comp))
        comp = np.isclose(proba, proba_post)
        self.assertFalse(np.any(comp))

        adj = [1.5, 0, -1.5]

        proba_post = prob_utils.adjust_proba(proba, adj)
        comp = proba_post < proba
        self.assertTrue(np.all([False, True] == comp[:, [0, 2]]))
        comp = proba_post > proba
        self.assertTrue(np.all([True, False] == comp[:, [0, 2]]))
        comp = np.isclose(proba, proba_post)
        self.assertFalse(np.all([False, True, False] == comp))
